#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
B-matrix calculation functions for MLM

@author: Joakim Linja
"""
from scipy.linalg import solve_triangular as soltri
from sklearn.utils.extmath import randomized_svd
from functools import partial
from numpy.random import normal as npnormal
from scipy.linalg import svd as spsvd
from numpy import (matrix as npmat, multiply as npmultiply,
                   eye as npeye, sqrt as npsqrt, finfo as npfinfo,
                   abs as npabs, var as npvar, transpose as nptranspose,
                   array as nparray, ndarray as npndarray, min as npmin, max as npmax)
from numpy.linalg import (qr as npqr, svd as npsvd, inv as npinv,
                          solve as npsolve, cholesky as npcholesky,
                          lstsq as nplstsq)



def _rankKrandomSVD(A: npmat, k: int):
    """
    https://amath.colorado.edu/faculty/martinss/Talks/2010_mmds_martinsson.pdf
    """
    _, n = A.shape
    # 1. O: n x k
    O = npnormal(size=(n,k))
    # 2. Y = A @ O, Y: m x k
    #Y = A @ O
    # 3. Q: m x k, Y = Q @ R
    Q, _ = npqr(A @ O, mode='reduced')#Y,mode='reduced')
    # 4. B = Q @ A, B: k x n
    #B = Q @ A
    # 5. B = U1 @ S @ Vh
    U1, S, Vh = npsvd(Q.T @ A, full_matrices=False)#B,full_matrices=False)
    # 6. U = Q @ U1
    return Q @ U1, S, Vh

#==============================================================================
#==============================================================================

def removeConstants(X,absolute=True):
    tX = type(X)
    if absolute:
        check = lambda y: npvar(y,axis=0)!=0
    else:
        check = lambda y: (npabs(npmax(y,axis=0)-npmin(y,axis=0))>=1e-5)

    if tX==npmat:
        mask = nparray(check(X))[0]
        X = (X.T[mask]).T
    elif tX==npndarray:
        mask = check(X)
        X = nptranspose(nptranspose(X)[mask])
    else:
        raise NotImplementedError
    return X, mask


def np_inverse(Dx: npmat, Dy: npmat):
    """
    B = ( Dx.T @ Dx)^-1 @ Dx.T @ Dy
    """
    return npinv(Dx.T @ Dx) @ Dx.T @ Dy


def np_inverse_dirty(Dx: npmat, Dy: npmat):
    """
    B = ( Dx.T @ Dx)^-1 @ Dx.T @ Dy
    """
    dig = npsqrt(npfinfo(float).eps)*npeye(Dx.shape[1])
    return npinv(Dx.T @ Dx + dig) @ Dx.T @ Dy


def np_solve(Dx: npmat, Dy: npmat):
    """
    (Dx.T @ Dx) @ B = Dx.T @ Dy
    """
    return npsolve(Dx.T @ Dx, Dx.T @ Dy)


def np_solve_dirty(Dx: npmat, Dy: npmat):
    """
    (Dx.T @ Dx) @ B = Dx.T @ Dy
    """
    dig = npsqrt(npfinfo(float).eps)*npeye(Dx.shape[1])
    return npsolve(Dx.T @ Dx + dig, Dx.T @ Dy)


def np_svd(Dx: npmat, Dy: npmat):
    """
    (U, S, V.T) = Dx
    B = V @ S^-1 @ U @ Dy
    """
    (U,S,Vh) = npsvd(Dx, full_matrices=False)
    return npmultiply(Vh.T, 1/S) @ U.T @ Dy


def sp_svd(Dx: npmat, Dy: npmat):
    """
    (U, S, V.T) = Dx
    B = V @ S^-1 @ U @ Dy
    """
    (U,S,Vh) = spsvd(Dx, full_matrices=False,compute_uv=True,
        overwrite_a=False,check_finite=False)
    return npmultiply(Vh.T, 1/S) @ U.T @ Dy


def rankKrandomSVD(p: float, Dx: npmat, Dy: npmat):
    """
    (U, S, V.T) ~= Dx
    B = V @ S^-1 @ U @ Dy
    """
    (U,S,Vh) = _rankKrandomSVD(Dx, int(Dx.shape[0]*p))
    return npmultiply(Vh.T, 1/S) @ U.T @ Dy


def sk_rand_svd(p: float, I, Dx, Dy):
    """
    (U, S, V.T) ~= Dx
    B = V @ S^-1 @ U @ Dy
    """
    U,S,Vh = randomized_svd(Dx,
                n_components=(int(Dx.shape[0]*p)),n_iter=I)
    return npmultiply(Vh.T, 1/S) @ U.T @ Dy


def cholesky_decomposition(Dx: npmat, Dy: npmat):
    """
    L @ L* = Dx.T @ Dx
    L @ y = I
    L* @ (Dx.T @ Dx)^-1 = y



    Dx = L @ L*
    Dx.T = (L*).T @ L.T
    Dx.T @ Dx = (L*).T @ L.T @ L @ L*
              = the same
    """
    if type(Dx)!=npmat:
        Dx = npmat(Dx)
    L = npcholesky(Dx.T @ Dx)
    imat = soltri(L.H, soltri(L, npeye(L.shape[1]),lower=True))
    return imat @ Dx.T @ Dy


def lstsq(Dx: npmat, Dy: npmat):
    """
    (Dx.T @ Dx + ~eps*I) @ B = Dx.T @ Dy
    """
    Dx = nparray(Dx)
    Dy = nparray(Dy)
    dig = npsqrt(npfinfo(float).eps)*npeye(Dx.shape[1])
    return nplstsq(Dx.T @ Dx + dig, Dx.T @ Dy, rcond=None)[0]


def lstsq_pure(Dx: npmat, Dy: npmat):
    """
    (Dx.T @ Dx) @ B = Dx.T @ Dy
    """
    return nplstsq(Dx.T @ Dx, Dx.T @ Dy, rcond=None)[0]


def getBcalcMethodDict():
    res = {'np_inverse':np_inverse,
           'np_inverse_dirty':np_inverse_dirty,
           'np_solve':np_solve,
           'np_solve_dirty':np_solve_dirty,
           'np_svd':np_svd,
           'sp_svd':sp_svd,
           'rankKrandomSVD_02':partial(rankKrandomSVD,0.2),
           'rankKrandomSVD_04':partial(rankKrandomSVD,0.4),
           'rankKrandomSVD_06':partial(rankKrandomSVD,0.6),
           'rankKrandomSVD_08':partial(rankKrandomSVD,0.8),
           'rankKrandomSVD_10':partial(rankKrandomSVD,1.0),
           'sk_rank_svd_02':partial(sk_rand_svd,0.2,1),
           'sk_rank_svd_04':partial(sk_rand_svd,0.4,1),
           'sk_rank_svd_06':partial(sk_rand_svd,0.6,1),
           'sk_rank_svd_08':partial(sk_rand_svd,0.8,1),
           'sk_rank_svd_10':partial(sk_rand_svd,1.0,1),
           'cholesky_decomposition':cholesky_decomposition,
           'lstsq':lstsq,
           'lstsq_pure':lstsq_pure}
    return res


def getBcalcMethodDict2():
    res = {'np_inverse':np_inverse,
           'np_inverse_dirty':np_inverse_dirty,
           'np_solve':np_solve,
           'np_solve_dirty':np_solve_dirty,
           'lstsq':lstsq,
           'lstsq_pure':lstsq_pure,
           'cholesky_decomposition':cholesky_decomposition,
           'np_svd':np_svd,
           'sp_svd':sp_svd,
           'rankKrandomSVD_10':partial(rankKrandomSVD,1.0),
           'rankKrandomSVD_08':partial(rankKrandomSVD,0.8),
           'rankKrandomSVD_06':partial(rankKrandomSVD,0.6),
           'rankKrandomSVD_04':partial(rankKrandomSVD,0.4),
           'rankKrandomSVD_02':partial(rankKrandomSVD,0.2),
           'sk_rank_svd_10':partial(sk_rand_svd,1.0,1),
           'sk_rank_svd_08':partial(sk_rand_svd,0.8,1),
           'sk_rank_svd_06':partial(sk_rand_svd,0.6,1),
           'sk_rank_svd_04':partial(sk_rand_svd,0.4,1),
           'sk_rank_svd_02':partial(sk_rand_svd,0.2,1)
           }
    return res
