#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Conversion functions from raw data into preprocessed format for MLM/EMLM usage

@author: Joakim Linja
"""
import numpy as np
import scipy.io.arff as arff
from pathlib import Path
from functools import wraps
from ML_functions import (preprocess,dobscv_SC,MinSorter2)
import physdata_functions as phdf 
import scipy.io as sio

_DatasetBasePth = Path('Datasets')


def arffconversion(func):
    """
    Conversion function for .arff type datasets
    """
    
    @wraps(func)
    def wrapper(unModOrig=False, noPreprocess=False):
        datasetPth, iN, oN = func()

        data, meta = arff.loadarff(datasetPth.as_posix())

        N = len(data)
        varNames = list(meta)
        varN = len(varNames)
        assert len(data[0]) == varN

        ndat = np.zeros((N,varN))
        for i in range(N):
            ndat[i,:] = np.array(list(data[i]))

        #return {'data':data,
        #        'arrdata':ndat,
        #        'meta':meta,
        #        'varNames':varNames,
        #        'inputN':iN,
        #        'outputN':oN}

        X = ndat[:,:iN]
        Y = ndat[:,iN:]
        if noPreprocess:
            return X, Y

        return preprocess(X,Y, unmodifiedOriginals=unModOrig)
    return wrapper





@arffconversion
def getSCM1D():
    """
    Dataset source: http://mulan.sourceforge.net/datasets-mtr.html
    Downloaded: 30.3.2020

    Observations: 9803
    Input features: 280
    Output features: 16

    All features are numeric
    """
    
    datasetPth = Path(_DatasetBasePth, 'mtr-datasets', 'scm1d.arff')
    return datasetPth, 280, 16


@arffconversion
def getATP1D():
    """
    Dataset source: http://mulan.sourceforge.net/datasets-mtr.html
    Downloaded: 30.3.2020

    Observations: 337
    Input features: 411
    Output features: 6

    All features are numeric
    """
    
    datasetPth = Path(_DatasetBasePth, 'mtr-datasets', 'atp1d.arff')
    return datasetPth, 411, 6


def getRF2(unModOrig=False, noPreprocess=False):
    """
    Dataset source: http://mulan.sourceforge.net/datasets-mtr.html
    Downloaded: 30.3.2020

    Observations: 9125 (7679)
    Input features: 576
    Output features: 8

    All features are numeric, partial observations removed.
    """
    
    datasetPth = Path(_DatasetBasePth, 'mtr-datasets', 'rf2.arff')

    data, meta = arff.loadarff(datasetPth.as_posix())

    data = map(lambda x: np.array(list(map(lambda y: float(y), x))), data)
    data = filter(lambda x: np.sum(np.isnan(x))==0, data)
    data = np.array(list(data))

    X = data[:,0:-8]
    Y = data[:,-8:]
    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def getIonosphere(unModOrig=False, noPreprocess=False):
    """
    Dataset source: https://archive.ics.uci.edu/ml/datasets/Ionosphere
    Downloaded 6.5.2020
    Classification

    Observations: 351
    Input features: 34
    Output features: 1

    All features are numeric
    """
    
    pth = Path(_DatasetBasePth,'Ionosphere/ionosphere.data')
    data = []
    with open(pth.as_posix(),'r') as f:
        data = f.readlines()

    def converter(x):
        try:
            return float(x)
        except ValueError:
            if x=='b':
                return 0
            elif x=='g':
                return 1
            else:
                raise NotImplementedError

    data = map(lambda x: x.strip(), data)
    data = map(lambda x: x.split(','), data)
    data = list(data)
    data = map(lambda x: list(map(lambda y: converter(y), x)), data)
    data = list(data)
    data = np.array(data)

    if noPreprocess:
        return data[:,:34], data[:,34].reshape(-1,1)

    return preprocess(data[:,:34],data[:,34].reshape(-1,1),
                      unmodifiedOriginals=unModOrig)


def getGisette(unModOrig=False,validationSet=False, noPreprocess=False):
    """
    Dataset source: https://archive.ics.uci.edu/ml/datasets/Gisette
    Downloaded 28.4.2020
    Classification

    Observations: 13500
        In use: 6000 (Gisette train set)
    Input features: 5000
    Output features: 1 (all values either 1 or -1)

    All features are numeric
    """
    
    pth = Path(_DatasetBasePth,'Gisette/gisette_train.data')
    pth2 = Path(_DatasetBasePth,'Gisette/gisette_train.labels')
    data = []
    with open(pth.as_posix(),'r') as f:
        data = f.readlines()

    data2 = []
    with open(pth2.as_posix(),'r') as f:
        data2 = f.readlines()

    def converter(dat):
        dat = map(lambda x: x.strip(), dat)
        dat = map(lambda x: x.split(' '), dat)
        dat = list(dat)
        dat = map(lambda x: list(map(lambda y: int(y), x)), dat)
        dat = list(dat)
        dat = np.array(dat)
        return dat

    X = converter(data)
    Y = converter(data2)

    if noPreprocess:
        return X, Y

    out = preprocess(X,Y, unmodifiedOriginals=unModOrig)
    if validationSet:
        pth = Path(_DatasetBasePth,'Gisette/gisette_valid.data')
        pth2 = Path(_DatasetBasePth,'Gisette/gisette_valid.labels')
        data = []
        with open(pth.as_posix(),'r') as f:
            data = f.readlines()

        data2 = []
        with open(pth2.as_posix(),'r') as f:
            data2 = f.readlines()

        vX = converter(data)
        vY = converter(data2)

        return (out,convertValidationSet(out,vX,vY))
    else:
        return out


def getMadelon(unModOrig=False,validationSet=False, noPreprocess=False):
    """
    Dataset source: https://archive.ics.uci.edu/ml/datasets/Madelon
    Downloaded 28.4.2020
    Classification

    Observations: 4400
        In use: 2000 (Madelon train set)
    Input features: 500
    Output features: 1 (all values either 1 or -1)

    All features are numeric
    """
    
    pth = Path(_DatasetBasePth,'Madelon/madelon_train.data')
    pth2 = Path(_DatasetBasePth,'Madelon/madelon_train.labels')
    data = []
    with open(pth.as_posix(),'r') as f:
        data = f.readlines()

    data2 = []
    with open(pth2.as_posix(),'r') as f:
        data2 = f.readlines()

    def converter(dat):
        dat = map(lambda x: x.strip(), dat)
        dat = map(lambda x: x.split(' '), dat)
        dat = list(dat)
        dat = map(lambda x: list(map(lambda y: int(y), x)), dat)
        dat = list(dat)
        dat = np.array(dat)
        return dat

    X = converter(data)
    Y = converter(data2)

    if noPreprocess:
        return X, Y

    out = preprocess(X,Y, unmodifiedOriginals=unModOrig)
    if validationSet:
        pth = Path(_DatasetBasePth,'Madelon/madelon_valid.data')
        pth2 = Path(_DatasetBasePth,'Madelon/madelon_valid.labels')
        data = []
        with open(pth.as_posix(),'r') as f:
            data = f.readlines()

        data2 = []
        with open(pth2.as_posix(),'r') as f:
            data2 = f.readlines()

        vX = converter(data)
        vY = converter(data2)

        return (out,convertValidationSet(out,vX,vY))
    else:
        return out


def getArtificialTestSet(N=1000, unModOrig=False, noPreprocess=False):
    """
    Observations: 1000 (default)
    Input features: 8
        True features: 4
        False features: 4
    Output features: 1

    All features are numeric
    """
    
    np.random.seed(seed=501)

    trueFeats = 4
    fakeFeats = 4#15

    x = np.linspace(-10,10,num=N)
    X = np.zeros((N,trueFeats+fakeFeats))

    X[:,0] = np.sin(2*x)
    X[:,1] = 2*np.cos(x)
    X[:,2] = (1+x)/5
    X[:,3] = (1/20)*np.power(x,2)
    #X[:,4:] = np.random.rand(N,fakeFeats)
    X[:,trueFeats:] = np.random.rand(N,fakeFeats)

    Y = (np.sum(X[:,:trueFeats],axis=1)).reshape(-1,1)

    (xn,xm) = np.shape(X)
    (yn,ym) = np.shape(Y)
    assert xn==yn
    assert xn==N
    assert xm==(trueFeats+fakeFeats)
    assert ym==1

    if noPreprocess:
        return X, Y

    return preprocess(X,Y)


def getAu38QT(unModOrig=False, noPreprocess=False):
    """
    Dataset source: research group Hannu Häkkinen

    Observations: 28761
    Input features: 763
    Output features: 1

    All features are numeric.
    """
    (X,Y) = phdf.mbtrDataExtraction()

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def getQM9Alu(unModOrig=False, noPreprocess=False):
    """
    jp6b10908_si_001 A+B+C+D+E
    All Aluminium
    """
    (X,Y) = phdf.QM9Alu(['A','B','C','D','E'])

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)
    
    
def getQM9AluA(unModOrig=False, noPreprocess=False):
    """
    jp6b10908_si_001 A
    All Aluminium
    """
    (X,Y) = phdf.QM9Alu(['A'])

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)

def getQM9AluB(unModOrig=False, noPreprocess=False):
    """
    jp6b10908_si_001 B
    All Aluminium
    """
    (X,Y) = phdf.QM9Alu(['B'])

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)

def getQM9AluC(unModOrig=False, noPreprocess=False):
    """
    jp6b10908_si_001 C
    All Aluminium
    """
    (X,Y) = phdf.QM9Alu(['C'])

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)

def getQM9AluD(unModOrig=False, noPreprocess=False):
    """
    jp6b10908_si_001 D
    All Aluminium
    """
    (X,Y) = phdf.QM9Alu(['D'])

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)

def getQM9AluE(unModOrig=False, noPreprocess=False):
    """
    jp6b10908_si_001 E
    All Aluminium
    """
    (X,Y) = phdf.QM9Alu(['E'])

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def getQM9fAlu(unModOrig=False, noPreprocess=False):
    """
    jp6b10908_si_001 A+B+C+D+E
    All Aluminium
    """
    (X,Y) = phdf.QM9Alu(['A','B','C','D','E'],full=True)

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)
    
    
def getQM9fAluA(unModOrig=False, noPreprocess=False):
    """
    jp6b10908_si_001 A
    All Aluminium
    """
    (X,Y) = phdf.QM9Alu(['A'],full=True)

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)

def getQM9fAluB(unModOrig=False, noPreprocess=False):
    """
    jp6b10908_si_001 B
    All Aluminium
    """
    (X,Y) = phdf.QM9Alu(['B'],full=True)

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)

def getQM9fAluC(unModOrig=False, noPreprocess=False):
    """
    jp6b10908_si_001 C
    All Aluminium
    """
    (X,Y) = phdf.QM9Alu(['C'],full=True)

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)

def getQM9fAluD(unModOrig=False, noPreprocess=False):
    """
    jp6b10908_si_001 D
    All Aluminium
    """
    (X,Y) = phdf.QM9Alu(['D'],full=True)

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)

def getQM9fAluE(unModOrig=False, noPreprocess=False):
    """
    jp6b10908_si_001 E
    All Aluminium
    """
    (X,Y) = phdf.QM9Alu(['E'],full=True)

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def SOAP_AluA(unModOrig=False, noPreprocess=False):
    """
    jp6b10908_si_001 A
    All Aluminium
    """
    (X,Y) = phdf.SOAP_Alu(['A'],full=True)

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM9(unModOrig=False, noPreprocess=False):
    """
    """
    (X,Y) = phdf.QM9()

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM7b(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/QM7')
    qm7bpth = Path(fldr,'qm7b.mat')
    qm7b = sio.loadmat(qm7bpth.as_posix())
    
    X = qm7b['X']
    Y = qm7b['T']
    
    (N,xi,xj) = np.shape(X)
    X1 = np.zeros((N,xi*xj))
    for k in range(N):
        for i in range(xi):
            for j in range(xj):
                X1[k,i*xi+xj] = X[k,i,j]
    
    if noPreprocess:
        return X1, Y
    return preprocess(X1, Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_Aspirin_5k(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'aspirin_dft.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    R = R[:5000,:,:]
    Y = Y[:5000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_Aspirin_10k(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'aspirin_dft.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    R = R[:10000,:,:]
    Y = Y[:10000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_Aspirin_15k(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'aspirin_dft.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    R = R[:15000,:,:]
    Y = Y[:15000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_Aspirin(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'aspirin_dft.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    #R = R[:15000,:,:]
    #Y = Y[:15000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_Azobenzene(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'azobenzene_dft.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    #R = R[:15000,:,:]
    #Y = Y[:15000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_Benzene2017(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'benzene2017_dft.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    #R = R[:15000,:,:]
    #Y = Y[:15000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_Benzene2018(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'benzene2018_dft.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    #R = R[:15000,:,:]
    #Y = Y[:15000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_Ethanol(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'ethanol_dft.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    #R = R[:15000,:,:]
    #Y = Y[:15000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)
        

def QM_sGDML_Malonaldehyde(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'azobenzene_dft.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    #R = R[:15000,:,:]
    #Y = Y[:15000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_Naphthalene(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'naphthalene_dft.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    #R = R[:15000,:,:]
    #Y = Y[:15000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_Paracetamol(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'paracetamol_dft.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    #R = R[:15000,:,:]
    #Y = Y[:15000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_Salicylic(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'salicylic_dft.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    #R = R[:15000,:,:]
    #Y = Y[:15000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_Toluene(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'toluene_dft.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    #R = R[:15000,:,:]
    #Y = Y[:15000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_Uracil(unModOrig=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'uracil_dft.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    #R = R[:15000,:,:]
    #Y = Y[:15000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_DHA(unModOrig=False, noPreprocess=False):
    """
    Docosahexaenoic acid
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'md22_DHA.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    #R = R[:15000,:,:]
    #Y = Y[:15000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def QM_sGDML_AlanineTetrapeptide(unModOrig=False, noPreprocess=False):
    """
    Ac-Ala3-NHMe
    """
    fldr = Path('Datasets/sGDML')
    dz = np.load(Path(fldr,'md22_Ac-Ala3-NHMe.npz'))
    Y = dz['E'].reshape((-1,1))
    #F = dz['F']
    R = dz['R']
    z = dz['z']
    
    #R = R[:15000,:,:]
    #Y = Y[:15000,:]
    X = phdf.sGDMLconversion(R,z,full=True)
    
    if noPreprocess:
        return X, Y
    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def getAu38QTfull(unModOrig=False, noPreprocess=False):
    """
    Dataset source: research group Hannu Häkkinen

    Observations: 28761
    Input features: 763
    Output features: 1

    All features are numeric.
    """
    (X,Y) = phdf.mbtrDataExtraction(full=True)

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def getAu38QTtoy(unModOrig=False, noPreprocess=False):
    """
    Dataset source: research group Hannu Häkkinen

    Observations: 28761
    Input features: 763
    Output features: 1

    All features are numeric.
    """
    (X,Y) = phdf.mbtrDataExtraction(k=2,N=5000,acc_N=20)

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def getAu38QTtoy10k(unModOrig=False, noPreprocess=False):
    """
    Dataset source: research group Hannu Häkkinen

    Observations: 28761
    Input features: 763
    Output features: 1

    All features are numeric.
    """
    (X,Y) = phdf.mbtrDataExtraction(k=2,N=10000,acc_N=20)

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)
    

def txt2Arr(pth: Path, splitter):
    """
    """
    dat = []
    with open(pth.as_posix(),'r') as f:
        dat = f.readlines()

    if len(splitter)==0:
        dat = map(lambda x: x.strip(), dat)
    else:
        dat = map(lambda x: x.strip().split(splitter), dat)
    dat = list(dat)
    dat = map(lambda x: list(map(lambda y: float(y), x)), dat)
    dat = list(dat)
    dat = np.array(dat)
    return dat


def getCOIL(unModOrig=False,validationSet=False, noPreprocess=False):
    """
    Dataset source:
        https://archive.ics.uci.edu/ml/datasets/Coil+1999+Competition+Data
    Classification

    Observations: 1800
    Input features: 20
    Output features: 1
    """
    fldr = Path('Datasets/COIL')
    tstXfp = Path(fldr,'COILTest.txt')
    trnXfp = Path(fldr,'COILTrain.txt')
    tstYfp = Path(fldr,'COILTestLabels.txt')
    trnYfp = Path(fldr,'COILTrainLabels.txt')
    X = txt2Arr(trnXfp,' ')
    Y = txt2Arr(trnYfp,' ')

    if noPreprocess:
        return X, Y

    out = preprocess(X,Y, unmodifiedOriginals=unModOrig)

    if validationSet:
        vX = txt2Arr(tstXfp,' ')
        vY = txt2Arr(tstYfp,' ')
        return (out,convertValidationSet(out,vX,vY))
    else:
        return out


def getCOIL2000(unModOrig=False,validationSet=False, noPreprocess=False):
    """
    Dataset source:
        https://archive.ics.uci.edu/ml/datasets/Insurance+Company+Benchmark+(COIL+2000)
    Classification

    Observations: 5220
    Input features: 85
    Output features: 1
    """
    fldr = Path('DataSets/COIL2000')
    datfp = Path(fldr,'ticdata2000.txt')
    evalfp = Path(fldr,'ticeval2000.txt')
    trgtfp = Path(fldr,'tictgts2000.txt')
    XY = txt2Arr(datfp,'\t')
    X = XY[:,0:-1]
    Y = XY[:,-1].reshape(-1,1)

    if noPreprocess:
        return X, Y

    out = preprocess(X,Y, unmodifiedOriginals=unModOrig)

    if validationSet:
        vX = txt2Arr(evalfp,'\t')
        vY = txt2Arr(trgtfp,'')
        return (out, convertValidationSet(out,vX,vY))
    else:
        return out


def getOptDigits(unModOrig=False,validationSet=False, noPreprocess=False):
    """
    Dataset source:
        https://archive.ics.uci.edu/ml/datasets/Optical+Recognition+of+Handwritten+Digits
    Classification

    Observations: 3823
    Input features: 62
    Output features: 1
    """
    fldr = Path('Datasets/OptDigits')
    #namesfp = Path(fldr,'optdigits.names')
    tesfp = Path(fldr,'optdigits.tes')
    trafp = Path(fldr,'optdigits.tra')
    XY = txt2Arr(trafp, ',')
    X = XY[:,0:-1]
    Y = XY[:,-1].reshape(-1,1)
    if noPreprocess:
        return X, Y

    out = preprocess(X,Y, unmodifiedOriginals=unModOrig)

    if validationSet:
        vXY = txt2Arr(tesfp,',')
        vX = vXY[:,0:-1]
        vY = vXY[:,-1].reshape(-1,1)
        return (out, convertValidationSet(out,vX,vY))
    else:
        return out


def getOutdoor(unModOrig=False,validationSet=False, noPreprocess=False):
    """
    Possibly https://archive.ics.uci.edu/ml/datasets/SML2010
    Classification

    Observations: 2400
    Input features: 21
    Output features: 1
    """
    fldr = Path('Datasets/Outdoor')
    tstXfp = Path(fldr,'OutDoorTest.txt')
    trnXfp = Path(fldr,'OutDoorTrain.txt')
    tstYfp = Path(fldr,'OutDoorTestLabels.txt')
    trnYfp = Path(fldr,'OutDoorTrainLabels.txt')
    X = txt2Arr(trnXfp,' ')
    Y = txt2Arr(trnYfp,' ')

    if noPreprocess:
        return X, Y


    out = preprocess(X,Y, unmodifiedOriginals=unModOrig)

    if validationSet:
        vX = txt2Arr(tstXfp,' ')
        vY = txt2Arr(tstYfp,' ')
        return (out, convertValidationSet(out,vX,vY))
    else:
        return out


def getPenDigits(unModOrig=False,validationSet=False, noPreprocess=False):
    """
    Dataaet source:
        https://archive.ics.uci.edu/ml/datasets/Pen-Based+Recognition+of+Handwritten+Digits
    Classification

    Observations: 7494
    Input features: 16
    Output features: 1
    """
    fldr = Path('Datasets/PenDigits')
    #namesfp = Path(fldr,'pendigits.names')
    tesfp = Path(fldr,'pendigits.tes')
    trafp = Path(fldr,'pendigits.tra')
    XY = txt2Arr(trafp, ',')
    X = XY[:,0:-1]
    Y = XY[:,-1].reshape(-1,1)

    if noPreprocess:
        return X, Y


    out = preprocess(X,Y, unmodifiedOriginals=unModOrig)

    if validationSet:
        vXY = txt2Arr(tesfp,',')
        vX = vXY[:,0:-1]
        vY = vXY[:,-1].reshape(-1,1)
        return (out, convertValidationSet(out,vX,vY))
    else:
        return out


def getSatImage(unModOrig=False,validationSet=False, noPreprocess=False):
    """
    https://archive.ics.uci.edu/ml/datasets/Statlog+(Landsat+Satellite)
    Classification

    Observations: 4435
    Input features: 36
    Output features: 1
    """
    fldr = Path('Datasets/')
    tstpth = Path(fldr,'statlog_satimage_test.dat')
    trnpth = Path(fldr,'statlog_satimage_train.dat')
    XY = txt2Arr(trnpth, ' ')
    X = XY[:,0:-1]
    Y = XY[:,-1].reshape(-1,1)
    Y[Y==7] = 6

    if noPreprocess:
        return X, Y

    out = preprocess(X,Y, unmodifiedOriginals=unModOrig)

    if validationSet:
        vXY = txt2Arr(tstpth,' ')
        vX = vXY[:,0:-1]
        vY = vXY[:,-1].reshape(-1,1)
        vY[vY==7] = 6
        return (out, convertValidationSet(out,vX,vY))
    else:
        return out


def getStudentTest(unModOrig=False,validationSet=False, noPreprocess=False):
    """
    Apparently: https://archive.ics.uci.edu/ml/datasets/student+performance
    But unknown preprocessing
    Classification

    Observations: 258
    Input features: 5
    Output features: 1
    """
    fldr = Path('Datasets/')
    tstpth = Path(fldr,'student_test.dat')
    trnpth = Path(fldr,'student_train.dat')
    XY = txt2Arr(trnpth, '\t')
    X = XY[:,0:-1]
    Y = XY[:,-1].reshape(-1,1)

    if noPreprocess:
        return X, Y


    out = preprocess(X,Y, unmodifiedOriginals=unModOrig)

    if validationSet:
        vXY = txt2Arr(tstpth,'\t')
        vX = vXY[:,0:-1]
        vY = vXY[:,-1].reshape(-1,1)
        return (out, convertValidationSet(out,vX,vY))
    else:
        return out


def getThyroidAnn(unModOrig=False,validationSet=False, noPreprocess=False):
    """
    https://archive.ics.uci.edu/ml/datasets/Thyroid+Disease
    Classification

    Observations: 3709
    Input features: 21
    Output features: 1
    """
    fldr = Path('Datasets/')
    tstpth = Path(fldr,'thyroid_ann_test.dat')
    trnpth = Path(fldr,'thyroid_ann_train.dat')
    XY = txt2Arr(trnpth, ' ')
    X = XY[:,0:-1]
    Y = XY[:,-1].reshape(-1,1)

    if noPreprocess:
        return X, Y

    out = preprocess(X,Y, unmodifiedOriginals=unModOrig)

    if validationSet:
        vXY = txt2Arr(tstpth,' ')
        vX = vXY[:,0:-1]
        vY = vXY[:,-1].reshape(-1,1)
        return (out, convertValidationSet(out,vX,vY))
    else:
        return out


def getIndoorLoc(unModOrig=False,validationSet=False, noPreprocess=False):
    """
    Dataset source: https://archive.ics.uci.edu/ml/datasets/UJIIndoorLoc
    Float regression / integer regression

    Observations: 19183
    Input features: 465
    Output features: 4
    """
    fldr = Path('Datasets/UJIndoorLoc')
    valpth = Path(fldr,'validationData.csv')
    trnpth = Path(fldr,'trainingData.csv')
    XY = np.genfromtxt(trnpth.as_posix(), delimiter=',')
    XY = XY[1:,:]
    # X is indices are 0-519
    # Y is indices are 520, 521, 522, 523
    # (longitude, langitude, floor [altitude], building ID)
    X = XY[:,0:520]
    Y = XY[:,520:524]

    if noPreprocess:
        return X, Y

    out = preprocess(X,Y, unmodifiedOriginals=unModOrig)

    if validationSet:
        vXY = np.genfromtxt(valpth.as_posix(), delimiter=',')
        vX = vXY[:,0:520]
        vY = vXY[:,520:524]
        return (out, convertValidationSet(out,vX,vY))
    else:
        return out


def getComputerActivity(unModOrig=False, noPreprocess=False):
    """
    Dataset source: http://www.cs.toronto.edu/~delve/data/comp-activ/desc.html
    Downloaded: 15.11.2019

    Observations: 8192
    Input features: 22
    Output features: 1
    """
    fldr = Path('Datasets/ComputerActivity/cpu')
    datfp = Path(fldr,'Prototask.data')
    XY = txt2Arr(datfp, ' ')
    X = XY[:,0:-1]
    Y = XY[:,-1].reshape(-1,1)

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def getCensus(unModOrig=False, noPreprocess=False):
    """
    Dataset source: http://www.cs.toronto.edu/~delve/data/census-house/desc.html
    Downloaded: 15.11.2019

    Observations: 22783
    Input features: 16
    Output features: 1
    """
    fldr = Path('Datasets/Census/house-price-16L')
    datfp = Path(fldr,'Prototask.data')
    XY = txt2Arr(datfp, ' ')
    X = XY[:,0:-1]
    Y = XY[:,-1].reshape(-1,1)

    if noPreprocess:
        return X, Y


    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def getMnist(unModOrig=False, validationSet=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets/Mnist')
    train = Path(fldr,'mnist.mat')
    test = Path(fldr,'mnist.t.mat')

    train = sio.loadmat(train.as_posix())
    test = sio.loadmat(test.as_posix())

    trainX = (train['Z']).toarray()
    trainY = np.array(train['y'])
    testX = test['Z'].toarray()
    testY = np.array(test['y'])

    W = 28*28
    (n1,w1) = np.shape(trainX)
    trainX = np.hstack((trainX, np.zeros((n1,W-w1))))
    (n1,w1) = np.shape(testX)
    testX = np.hstack((testX, np.zeros((n1,W-w1))))

    X = trainX
    Y = trainY

    if noPreprocess:
        return X, Y

    out = preprocess(X,Y, unmodifiedOriginals=unModOrig)

    if validationSet:
        vX = testX
        vY = testY
        return (out, convertValidationSet(out,vX,vY))
    else:
        return out


def getAu38Q(sel, unModOrig=False, validationSet=False, noPreprocess=False):
    """

    sel:

        0. 12413x4000
        1. 12413x400
        2. 12413x80
        3. 8000x4000
        4. 8000x400
        5. 8000x80
        6. 4000x4000
        7. 4000x400
        8. 4000x80
    """
    fldr = Path('Datasets/Au38Q_MBTR-K3')
    vers = {
        0:'N100-12413.npy',
        1:'N100-8000.npy',
        2:'N100-4000.npy',
        3:'N10-12413.npy',
        4:'N10-8000.npy',
        5:'N10-4000.npy',
        6:'N2-12413.npy',
        7:'N2-8000.npy',
        8:'N2-4000.npy',
        }
    nm = 'Au38Q_MBTR-K3-'
    pth = Path(fldr,nm+vers[sel])

    XY = np.load(pth.as_posix())

    X = XY[:,0:-1]
    Y = XY[:,-1].reshape(-1,1)

    if noPreprocess:
        return X, Y

    return preprocess(X,Y, unmodifiedOriginals=unModOrig)


def Au38QT_orig(unModOrig=False, validationSet=False, noPreprocess=False):
    """
    """
    fldr = Path('Datasets','19-07-10-MBTR-K2-Preprocessed-MATLAB-Au38-Extended')
    test = sio.loadmat(Path(fldr,'Test.mat').as_posix())
    train = sio.loadmat(Path(fldr,'Train.mat').as_posix())

    trainX = train['Xtrain']
    trainY = train['Ytrain']
    testX = test['Xtest']
    testY = test['Ytest']

    if noPreprocess:
        if validationSet:
            return ((trainX,trainY),(testX,testY))
        else:
            return (trainX,trainY)

    out = preprocess(trainX,trainY, unmodifiedOriginals=unModOrig)
    if validationSet:
        val = convertValidationSet(out,testX,testY)
        return (out,val)
    else:
        return out


def Au38QT_orig2(unModOrig=False, validationSet=False, noPreprocess=False):
    """
    """
    
    fldr = Path('Datasets','19-07-10-MBTR-K2-Preprocessed-MATLAB-Au38-Extended')
    test = sio.loadmat(Path(fldr,'Test.mat').as_posix())
    train = sio.loadmat(Path(fldr,'Train.mat').as_posix())

    trainX = train['Xtrain']
    trainY = train['Ytrain']
    testX = test['Xtest']
    testY = test['Ytest']

    if noPreprocess:
        if validationSet:
            return ((trainX,trainY),(testX,testY))
        else:
            return (trainX,trainY)

    out = preprocess(trainX,trainY, unmodifiedOriginals=unModOrig,
                     Yinterval=[-1,1])
    if validationSet:
        val = convertValidationSet(out,testX,testY)
        return (out,val)
    else:
        return out
    

def convertValidationSet(trainOut,valX,valY):
    """
    """
    (X, Y, nX, nY, Xscaler, Yscaler, XconstFiltMask, YconstFiltMask) = trainOut
    vX = valX[:,XconstFiltMask]
    vX = Xscaler.convertToScale(vX)
    vY = valY[:,YconstFiltMask]
    vY = Yscaler.convertToScale(vY)
    return (vX,vY)


def savePreprocessedDatasets(dskeys):
    """
    """
    datasets = getDatasets()

    #dsf = Path('Experiments/Saved_datasets')
    #parf = Path(dsf,'Params')
    #dsf = Path('Experiments/Saved_datasets')
    dsf = Path('CData/Saved_datasets')
    parf = Path(dsf,'Params')

    def fpth(pth,add):
        return Path(pth,add).as_posix()
    
    #fd = getDatasetsWithValidationSets()

    for dskey in dskeys:
        print(dskey)

        df = datasets[dskey]
        res = df(unModOrig=True)
        X, Y, nX, nY, Xscaler, Yscaler, XconstFiltMask, YconstFiltMask = res

        
        np.save(fpth(dsf,'{0:s}_origX.npy').format(dskey),X)
        np.save(fpth(dsf,'{0:s}_origY.npy').format(dskey),Y)
        np.save(fpth(dsf,'{0:s}_normX.npy').format(dskey),nX)
        np.save(fpth(dsf,'{0:s}_normY.npy').format(dskey),nY)
        (xmin,xmax) = Xscaler.getScalingParams()
        (ymin,ymax) = Yscaler.getScalingParams()
        np.save(fpth(parf,'{0:s}_xmin.npy').format(dskey),xmin)
        np.save(fpth(parf,'{0:s}_xmax.npy').format(dskey),xmax)
        np.save(fpth(parf,'{0:s}_ymin.npy').format(dskey),ymin)
        np.save(fpth(parf,'{0:s}_ymax.npy').format(dskey),ymax)
        np.save(fpth(parf,'{0:s}_xconstmask.npy').format(dskey),XconstFiltMask)
        np.save(fpth(parf,'{0:s}_yconstmask.npy').format(dskey),YconstFiltMask)


def getDatasets():
    return {'SCM1D':getSCM1D,
            'ATP1D':getATP1D,
            'RF2':getRF2,
            #'Ionosphere':getIonosphere,
            #'Gisette':getGisette,
            #'Madelon':getMadelon,
            #'Artificial':getArtificialTestSet,
            #'Au38QT':getAu38QT,
            #'COIL':getCOIL,
            #'COIL2000':getCOIL2000,
            #'OptDigits':getOptDigits,
            #'Outdoor':getOutdoor,
            #'PenDigits':getPenDigits,
            #'SatImage':getSatImage,
            #'StudentTest':getStudentTest,
            #'ThyroidAnn':getThyroidAnn,
            #'IndoorLoc':getIndoorLoc,
            'ComputerActivity':getComputerActivity,
            'Census':getCensus,
            #'Mnist':getMnist,
            #'Au38Q-N100-12k':partial(getAu38Q,0),
            #'Au38Q-N100-8k':partial(getAu38Q,1),
            #'Au38Q-N100-4k':partial(getAu38Q,2),
            #'Au38Q-N10-12k':partial(getAu38Q,3),
            #'Au38Q-N10-8k':partial(getAu38Q,4),
            #'Au38Q-N10-4k':partial(getAu38Q,5),
            #'Au38Q-N2-12k':partial(getAu38Q,6),
            #'Au38Q-N2-8k':partial(getAu38Q,7),
            #'Au38Q-N2-4k':partial(getAu38Q,8),
            #'Au38QT-orig':Au38QT_orig,
            #'Au38QT-orig2':Au38QT_orig2,
            #'Al100_n3l1':Al100_n3l1,
            #'Al100_n6l1':Al100_n6l1,
            #'Al100_n6l4':Al100_n6l4,
            #'QM9Alu':getQM9Alu,
            #'QM9AluA':getQM9AluA,
            #'QM9AluB':getQM9AluB,
            #'QM9AluC':getQM9AluC,
            #'QM9AluD':getQM9AluD,
            #'QM9AluE':getQM9AluE,
            #'QM9fAlu':getQM9fAlu,
            #'QM9fAluA':getQM9fAluA,
            #'QM9fAluB':getQM9fAluB,
            #'QM9fAluC':getQM9fAluC,
            #'QM9fAluD':getQM9fAluD,
            #'QM9fAluE':getQM9fAluE,
            #'Aspirin_5k':QM_sGDML_Aspirin_5k,
            #'Aspirin_10k':QM_sGDML_Aspirin_10k,
            #'Aspirin_15k':QM_sGDML_Aspirin_15k,
            #'Aspirin':QM_sGDML_Aspirin,
            #'Azobenzene':QM_sGDML_Azobenzene,
            #'Benzene2017':QM_sGDML_Benzene2017,
            #'Benzene2018':QM_sGDML_Benzene2018,
            #'Ethanol':QM_sGDML_Ethanol,
            #'Malonaldehyde':QM_sGDML_Malonaldehyde,
            #'Naphthalene':QM_sGDML_Naphthalene,
            #'Paracetamol':QM_sGDML_Paracetamol,
            #'Salicylic':QM_sGDML_Salicylic,
            #'Toluene':QM_sGDML_Toluene,
            #'Uracil':QM_sGDML_Uracil,
            #'DocosahexaenoicAcid':QM_sGDML_DHA,
            #'AlanineTetrapeptide':QM_sGDML_AlanineTetrapeptide,
            #'Au38QTfull':getAu38QTfull,
            #'Au38K2toy':getAu38QTtoy,
            #'Au38K2toy10k':getAu38QTtoy10k,
            }

def getDatasetsWithValidationSets():
    return {
            'Gisette':getGisette,
            'Madelon':getMadelon,
            'COIL':getCOIL,
            'COIL2000':getCOIL2000,
            'OptDigits':getOptDigits,
            'Outdoor':getOutdoor,
            'PenDigits':getPenDigits,
            'SatImage':getSatImage,
            'StudentTest':getStudentTest,
            'ThyroidAnn':getThyroidAnn,
            'IndoorLoc':getIndoorLoc,
            'Mnist':getMnist,
            'Au38QT-orig':Au38QT_orig,
            'Au38QT-orig2':Au38QT_orig2,
            }


def processValidationSets():
    fd = getDatasetsWithValidationSets()
    #dsf = Path('Experiments/Saved_datasets')
    dsf = Path('Data/Saved_datasets')
    parf = Path(dsf,'Params')
    def fpth(pth,add):
        return Path(pth,add).as_posix()

    for dkey in fd:
        print('DsKey:',dkey)
        (out,(vX,vY)) = fd[dkey](unModOrig=True,validationSet=True)
        np.save(fpth(dsf,'{0:s}_valX.npy').format(dkey),vX)
        np.save(fpth(dsf,'{0:s}_valY.npy').format(dkey),vY)



def getArt10(unModOrig=False):
    return getArtificialTestSet(N=10,unModOrig=unModOrig)

def getArt50(unModOrig=False):
    return getArtificialTestSet(N=50,unModOrig=unModOrig)

def getArt100(unModOrig=False):
    return getArtificialTestSet(N=100,unModOrig=unModOrig)

def getArt500(unModOrig=False):
    return getArtificialTestSet(N=500,unModOrig=unModOrig)

def getArt1000(unModOrig=False):
    return getArtificialTestSet(N=1000,unModOrig=unModOrig)

def getArt5000(unModOrig=False):
    return getArtificialTestSet(N=5000,unModOrig=unModOrig)


def getDatasets2():
    return {'Artificial0010':getArt10,
            'Artificial0050':getArt50,
            'Artificial0100':getArt100,
            'Artificial0500':getArt500,
            'Artificial1000':getArt1000,
            'Artificial5000':getArt5000,
            }


def TrainValPreSplit(dkey, X, Y, xscaler, yscaler, XconstFiltMask, outputSel=-1,
                 separateValidationSet=False,
                 validationSetP=0.33,validationSet=None,
                 deterministicDobSCV=False,
                 forceTrainValSplit=True):
    """
    Includes the ability to force own train/val split
    """
    print('TrainValPreSplit')

    oX = xscaler.convertFromScale(X)
    oY = yscaler.convertFromScale(Y)

    (N,xm) = np.shape(oX)
    #valXconstFiltMask = None
    #valYconstFiltMask = None

    if outputSel != -1:
        oY = oY[:,outputSel]
        oY = oY.reshape((N,1))

    # Dataset train/Val presplit number
    for itrn in range(0,30):

        if type(validationSet) is tuple and not forceTrainValSplit:
            raise NotImplementedError

        else:
            print('Taking validation dataset from training data with FoldSorter')
            if type(validationSet) is tuple and forceTrainValSplit:
                print('There is a readily provided validation dataset.')
                print('Combining it with the training data.')
                (valX, valY) = validationSet
                #valN = len(valX)
                #print(valY)

                """
                validation set converter:
                (X, Y, nX, nY, Xscaler, Yscaler, XconstFiltMask, YconstFiltMask) = trainOut
                vX = valX[:,XconstFiltMask]
                vX = Xscaler.convertToScale(vX)
                vY = valY[:,YconstFiltMask]
                vY = Yscaler.convertToScale(vY)
                """

                #ovalX = xscaler.convertFromScale(valX)
                ovalY = yscaler.convertFromScale(valY)

                # The training/validation sets combined:
                #ocX = np.vstack((oX,ovalX))
                ocY = np.vstack((oY,ovalY))
            else:
                # If forcing is not done, rename
                #ocX = oX
                ocY = oY

            # Separate ocX and ocY into nF separate folds
            nF = int(np.round(1/validationSetP))
            #print('Total N:',len(ocY))
            if deterministicDobSCV:
                raise NotImplementedError
                #(Folds, rInds) = FoldSortMin(ocY[:],nF)
                (valInds,valX), (trnInds,trnX) = MinSorter2(ocY[:],nF,sel=-2)

            else:
                (Folds, rInds) = dobscv_SC(ocY[:],nF,deterministic=False)

                valInds = rInds[0]
                trnInds = rInds[1]
                for i in range(2,len(rInds)):
                    trnInds = np.hstack((trnInds,rInds[i]))


            np.save('CData/Saved_datasets/Splits/{0:s}_valInds_o{2:d}_itrn{1:d}.npy'.format(
                dkey,itrn,outputSel),valInds)
            np.save('CData/Saved_datasets/Splits/{0:s}_trnInds_o{2:d}_itrn{1:d}.npy'.format(
                dkey,itrn,outputSel),trnInds)


def runRSmaximin(dskeys):
    """
    Run full RS-maximin to the datasets and save the index list
    """
    from mlm import rps
    dfldr = Path('Datasets/Saved_datasets')
    
    for dskey in dskeys:
        
        nX = np.load(Path(dfldr,dskey+'_normX.npy'))
        nY = np.load(Path(dfldr,dskey+'_normY.npy'))
        
        R, T, I = rps(nX,nY,100, 'RS-maximin', returnI=True,
                forceCalculation=True,
                progressingI=True,saveInterval=10000)
        
        np.save('CData/Saved_datasets/RTI/{0:s}_R.npy'.format(dskey),R)
        np.save('CData/Saved_datasets/RTI/{0:s}_T.npy'.format(dskey),T)
        np.save('CData/Saved_datasets/RTI/{0:s}_I.npy'.format(dskey),I)


if __name__ == "__main__":
    
    #QM9()
    #"""
    datasets = getDatasets()

    def sizePrinter(key):
        print(key)
        df = getDatasets()
        X, Y = (df[key])(noPreprocess=True)
        print('shape(X):',np.shape(X))
        print('shape(Y):',np.shape(Y))
        print('')


    datasetKeys = list((getDatasets()).keys())
    #"""
    #for datasetKey in datasetKeys:
    #    sizePrinter(datasetKey)#"""


    savePreprocessedDatasets(datasetKeys)
    #runRSmaximin(datasetKeys)

    #processValidationSets()
    
    #QM_sGDML_Aspirin(unModOrig=False, noPreprocess=False)
    

    print('\nEnd of Line')
