#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Some disk operations

@author: Joakim Linja
"""
from pathlib import Path
import pathlib
from os import remove


def getFilePaths(folderPath: Path, filesuffix: str, includeSubfolders=True):
    """
    Traverses a given folder and its subfolders etc recursively and
    forms a list of all file addresses that match a given suffix
    
    Parameters
    ----------
    *folderPath* : Path-object
        Contains the file path of the target folder to be looked through
        
    *filesuffix* : str
        The filesuffix for the files to be searched for
        
    *includeSubfolders* : boolean, optional
        Determines whether subfolders are also searched for files with filesuffix
    
        Default: True
    
    Returns
    -------
    *res* : The list of filepaths that were found
    """
    
    filePaths = []

    # Function definitions
    leaveFiles = lambda x: x.is_file()
    leaveDirs = lambda x: x.is_dir()

    getPathsWithFilter = lambda fFun,fPath: filter(fFun,fPath.iterdir())

    # Locate files
    files = getPathsWithFilter(leaveFiles,folderPath)

    # Form the string which is matched against
    mtch = '*{0}'.format(filesuffix)

    # Get all files in this folder that match the file suffix
    filePaths.extend(list(filter(lambda x: x.match(mtch),files)))

    # Get names of all subfolders
    subdirs = list(getPathsWithFilter(leaveDirs,folderPath))

    if includeSubfolders:
        # Repeat getFilePaths in each found subfolder
        for subdir in subdirs:
            subres = getFilePaths(subdir, filesuffix,
                      includeSubfolders=includeSubfolders)
            filePaths.extend(subres)

    # filePaths now contains all files which had filesuffix-suffix and were 
    # found in the given folder and subfolders (if includeSubFolders is True)

    # At the end, filter out wrong and empty filepaths
    res = list(filter(lambda x: type(x)==pathlib.PosixPath,filePaths))
    return res


def removeFile(filepath: Path,suffix: str):
    """
    Remove a file with the given filepath and file suffix.
    Do nothing if filepath is wrong or the suffix is wrong.
    (it is a safety precaution against the user)
    
    Parameters
    ----------
    *filepath* : Path-object
        Path of the file to-be-removed. 
    
    *suffix* : str
        The suffix of the to-be-removed file. This is mainly an error reduction 
        ensurance. 
    
    """
    
    if filepath.exists():
        if filepath.match('*{0}'.format(suffix)):
            remove(filepath)
