#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
A collection of feature selection functions

@author: Joakim Linja
@author: Tommi Kärkkäinen
"""
import numpy as np
from matlabLike import bsxfun
import sys
import psutil
from mlm import rps
import pandas as pd
from autofeat import FeatureSelector
from time import process_time, time
import emlm
from kneed import KneeLocator
from copy import deepcopy
from ML_functions import RMSE
from pathlib import Path


def _timestamp(add=None):
    if add is not None:
        return (process_time(), time(), add)
    return (process_time(), time())


def MasFS(X: np.ndarray, Xk: np.ndarray,
          W: np.ndarray, meanMAS=False):
    """
    Matlab implementation doctext:
        UNTITLED2 MAS (Mean Absolute Sensitivity) type feature relevance
        % assessment for the Extreme Minimal Learning Machine
        %   (c) Tommi K. and Monzuun Oy, May 23, 2019
        
    Parameters
    ----------
    *X* : numpy.ndarray
        Input dataset
    
    *Xk* : numpy.ndarray
        Reference points selected from *X*
        
    *W* : numpy.ndarray
        Weight matrix from EMLM model
        
    *meanMAS* : boolean, optional
        Set true to use original technique by Tommi K
        
        Default: False
        
    Returns
    -------
    *WDH* : numpy.ndarray
        MAS values for each feature of *X*
    """
    (N, n) = np.shape(X)
    (k, _) = np.shape(W)
    WDH = np.zeros((k,n))

    eps = np.finfo(float).eps

    # for i=1:N
    for i in range(N):
        # D = bsxfun(@minus, Xk, X(i,:)); %(m,n)
        D = bsxfun(lambda a,b: a-b, Xk, X[i,:])
        # l2n = sqrt(sum(D.^2,2)); %(m,1)
        l2n = np.sqrt(np.sum(np.square(D), axis=1))
        # D = bsxfun(@rdivide, D, max(eps,l2n)); %(m,n)
        D = D / np.max(eps, l2n)
        # WDH = WDH + abs(W*D); %(k,n), AS - Absolute Sensitivity
        WDH += np.abs(W @ D)
    WDH = WDH/N

    if meanMAS:
        # Original technique from (TK ESANN 2015), i.e., mean of MAS
        return np.mean(WDH,axis=0) #0 or 1 ?
    else:
        # Only last MAS values related to minority class
        return WDH[k,:]


def MasFSreg(X: np.ndarray, Xk: np.ndarray, W: np.ndarray, printT=False):
    """
    Original doctext:
        % UNTITLED2 MAS (Mean Absolute Sensitivity) type feature relevance
        % assessment for the Extreme Minimal Learning Machine
        %   (c) Tommi K. and Monzuun Oy, May 23, 2019

    MasFS for regression

    Parameters
    ----------
    *X* : numpy.ndarray
        Input dataset
    
    *Xk* : numpy.ndarray
        Reference points selected from *X*
        
    *W* : numpy.ndarray
        Weight matrix from EMLM model
        
    *printT* : boolean, optional
        Print progress to console
        
        Default: False
        
    Returns
    -------
    *MASvec* : numpy.ndarray
        MAS values for each feature of *X*
    """
   
    (N, n) = np.shape(X)
    MASvec = np.zeros((1,n))
    cof = 1/N
    eps = np.finfo(float).eps

    # for i=1:N
    for i in range(N):
        if printT:
            msg = 'MasFSreg: {0:d} / {1:d} (cpu {2:.1f}%)'.format(
                    i+1,N,psutil.cpu_percent())
            sys.stdout.write('\r{0}'.format(msg))
        # D = bsxfun(@minus, Xk, X(i,:)); %(m,n)
        D = Xk-X[i,:]
        #D = bsxfun(lambda a,b: a-b, Xk, X[i,:])

        # l2n = sqrt(sum(D.^2,2)); %(m,1)
        l2n = np.sqrt(np.sum(np.square(D),axis=1))

        # D = bsxfun(@rdivide, D, max(eps,l2n)); %(m,n)
        Nl2n = len(l2n)
        maxes = np.max(np.vstack((np.repeat(eps,Nl2n),l2n)), axis=0)
        D = D / maxes.reshape(-1,1)

        # MASvec = MASvec + cof*abs(W*D); %(1,n)
        MASvec += cof*np.abs(W @ D)
        
    if printT:
        sys.stdout.write('\n')

    return MASvec


def MasFSregCityBlock(X: np.ndarray, Xk: np.ndarray, W: np.ndarray, printT=False):
    """
    Original doctext:
        %UNTITLED2 MAS (Mean Absolute Sensitivity) type feature relevance
        % assessment for the Extreme Minimal Learning Machine
        %   (c) Tommi K., May 23, 2019
        % Inputs
        % - X: (N,n) matrix of learning data
        % - Xk: (m,n) matrix of reference points
        % - W: (1,m) matrix of distance-regression weights

    Cityblock distance
    """
   
    (N, n) = np.shape(X)
    MASvec = np.zeros((1,n))
    cof = 1/N
    eps = np.finfo(float).eps

    # for i=1:N
    for i in range(N):
        if printT:
            msg = 'MasFSregEucl: {0:d} / {1:d} (cpu {2:.1f}%)'.format(
                    i+1,N,psutil.cpu_percent())
            sys.stdout.write('\r{0}'.format(msg))
        # D = bsxfun(@minus, Xk, X(i,:)); %(m,n)
        D = Xk-X[i,:]
        #D = bsxfun(lambda a,b: a-b, Xk, X[i,:])

        # l2n = sqrt(sum(D.^2,2)); %(m,1)
        l2n = np.sqrt(np.sum(np.square(D),axis=1))
        l2n = np.sum(np.abs(D),axis=1)
        #print('shape(l2n)',np.shape(l2n))

        # D = bsxfun(@rdivide, D, max(eps,l2n)); %(m,n)
        Nl2n = len(l2n)
        maxes = np.max(np.vstack((np.repeat(eps,Nl2n),l2n)), axis=0)
        D = D / maxes.reshape(-1,1)

        # MASvec = MASvec + cof*abs(W*D); %(1,n)
        #MASvec += cof*np.abs(W @ D)
        MASvec += cof*np.abs(W @ D)
    #MASvec *= cof
    if printT:
        sys.stdout.write('\n')
    return MASvec


def histfun(data,bincount=10):
    """
    This is mainly to check if this is how the readily implemented histogram
    works. Result is that this is equivalent to numpy.histogram(data,bins=10)

    Matlab's histogram does something else. It does not return same results.
    """
    maxi = bincount-1
    shp = np.shape(data)
    if len(shp)==2:
        if shp[0]>shp[1]:
            dat = data[:,0]
        else:
            dat = data[0,:]
    elif len(shp)>2:
        raise NotImplementedError
    else:
        dat = data
    mi = np.min(dat)
    ma = np.max(dat)
    edges = np.linspace(mi,ma,num=(bincount+1),endpoint=True)
    binC = np.zeros((bincount,),dtype=np.int)
    for i in range(bincount):
        l = edges[i]
        u = edges[i+1]
        if i==maxi:
            m = np.logical_and(dat >= l, dat <= u)
        else:
            m = np.logical_and(dat >= l, dat < u)
        dt = dat[m]
        binC[i] = len(dt)
    return binC, edges


def autofeatsel(X, Y, regression=1):
    """
    """
    fsel = FeatureSelector(verbose=1)
    (h,w) = np.shape(X)
    cols = list(map(lambda i: 'X{0:d}'.format(i), range(w)))
    nX = fsel.fit_transform(pd.DataFrame(X,columns=cols))
    nX = np.to_numpy()
    return nX


def kneePointFS(rvals, inds):
    # Kneepoint
    if len(rvals)==1:
        rankordr = np.argsort(rvals[0])
        rvals = rvals[0]
    else:
        rankordr = np.argsort(rvals)
    
    orderedRVals = np.flip(rvals[rankordr])
    # orderedRVals is now best to worst order
    orderedRValsX = np.arange(len(orderedRVals))
    
    
    # Kneepoint
    kneedle = KneeLocator(orderedRValsX,orderedRVals,curve='convex',direction='decreasing',
                          interp_method="interp1d", online=True)
    
    kneeval = kneedle.knee_y 
    # Figure out the true index of the kneepoint
    kneeInd = np.where(kneeval == rvals)[0][0]
    kneeIndOrdr = np.where(kneeInd == rankordr)[0][0]
    
    toBeRemovedInds = rankordr[:kneeIndOrdr]

    Nf = len(rankordr)
    
    toBeKeptBool = np.ones(Nf,dtype=bool)
    toBeKeptBool[toBeRemovedInds] = False

    tind = deepcopy(inds)
    

    tt = tind[tind]

    tind[tind] = np.logical_and(tt, toBeKeptBool)
    
    featureInds = np.logical_and(
            inds,tind)
    
    return featureInds


def FS_MASknee(inds, sX, nY, vX, vY, sYsclr, 
                 forceTrainValSplit=False,
                 regression=True, savekey='', savenum=-1,
                 savefldr='',refP=85, resfldr='',
                 datakey='',outputSel=-1,ite=0, full=False,
                 namePrefix=None):
    """
    A single function version of the MASvec-kneepoint feature selection 
    algorithm.
    """
    print('FS_MASknee')
    
    if full:
        refp = 100
    else:
        refp = refP
    
    print('refp:',refp)
    
    if namePrefix is None:
        nmPrfx = 'Reference_RA_'
    else:
        nmPrfx = ''
    
    
    featpth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'MASknee_Feats',
           refp,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    errpth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'MASknee_Err',
           refp,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    rankpth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'MASknee_featR',
           refp,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    timepth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'MASknee_Times',
           refp,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    
    
    tsX = sX[:,inds]
    tvX = vX[:,inds]
    (tsXi,tsXj) = np.shape(tsX)
    (tvXi,tvXj) = np.shape(tvX)
    
    
    (pt0,wt0) = _timestamp() # Train model
    
    W, R = emlm.train(tsX,nY,beta=np.sqrt(np.finfo(float).eps),
              K_rel=refp,rs_method="RS-maximin", method='svd', weights=None)
    
    (pt1,wt1) = _timestamp()# calculate MAS
    
    rvals = MasFSreg(tsX, R, W, printT=False)
    
    print('shape(rvals)',np.shape(rvals))
    print(rvals)
    
    (pt2,wt2) = _timestamp()
    
    # Kneepoint
    if len(rvals)==1:
        rankordr = np.argsort(rvals[0])
        rvals = rvals[0]
    else:
        rankordr = np.argsort(rvals)
    
    orderedRVals = np.flip(rvals[rankordr])
    # orderedRVals is now best to worst order
    orderedRValsX = np.arange(len(orderedRVals))
    
    (pt3,wt3) = _timestamp()
    
    # Kneepoint
    kneedle = KneeLocator(orderedRValsX,orderedRVals,curve='convex',direction='decreasing',
                          interp_method="interp1d", online=True)
    
    (pt4,wt4) = _timestamp()
    
    kneeval = kneedle.knee_y 
    # Figure out the true index of the kneepoint
    kneeInd = np.where(kneeval == rvals)[0][0]
    kneeIndOrdr = np.where(kneeInd == rankordr)[0][0]
    
    toBeRemovedInds = rankordr[:kneeIndOrdr]

    Nf = len(rankordr)
    
    toBeKeptBool = np.ones(Nf,dtype=bool)
    toBeKeptBool[toBeRemovedInds] = False

    tind = deepcopy(inds)
    

    tt = tind[tind]

    tind[tind] = np.logical_and(tt, toBeKeptBool)

    featureInds = np.logical_and(
            inds,tind)
    
    tsX = sX[:,featureInds]
    tvX = vX[:,featureInds]
    
    (pt5,wt5) = _timestamp()
    
    W, R = emlm.train(tsX,nY,beta=np.sqrt(np.finfo(float).eps),
              K_rel=refp,rs_method="RS-maximin", method='svd', weights=None)
    
    (pt6,wt6) = _timestamp()
    
    predY = emlm.pred2(tvX, W, R)
    
    (pt7,wt7) = _timestamp()
    
    valErr = RMSE(vY,predY.reshape((-1,1)))
    
    (pt8,wt8) = _timestamp()
    
    """
      train   MAS     ordr    knecre  stuff   train   pred    rmse
    [[pt1-pt0,pt2-pt1,pt3-pt2,pt4-pt3,pt5-pt4,pt6-pt5,pt7-pt6,pt8-pt7,
      wt1-wt0,wt2-wt1,wt3-wt2,wt4-wt3,wt5-wt4,wt6-wt5,wt7-wt6,wt8-wt7]]
    """

    times = [pt1-pt0,pt2-pt1,pt3-pt2,pt4-pt3,pt5-pt4,pt6-pt5,pt7-pt6,pt8-pt7,
             wt1-wt0,wt2-wt1,wt3-wt2,wt4-wt3,wt5-wt4,wt6-wt5,wt7-wt6,wt8-wt7]
    
    np.save(featpth,featureInds)
    np.save(errpth,valErr)
    np.save(rankpth,rvals)
    np.save(timepth,times)
    

def FS_MASkneeFull(inds, sX, nY, vX, vY, sYsclr, 
                 forceTrainValSplit=False,
                 regression=True, savekey='', savenum=-1,
                 savefldr='',refP=85, resfldr='',
                 datakey='',outputSel=-1,ite=0, full=True):
    """
    A single function version of the MASvec-kneepoint feature selection 
    algorithm.
    """
    if full:
        refp = 100
    else:
        raise NotImplementedError 
        refp = refP
    
    errpth = Path(resfldr,
       'Reference_RA_{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'MASknee_Err',
           refp,
           ite, forceTrainValSplit, outputSel))
    rankpth = Path(resfldr,
       'Reference_RA_{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'MASknee_featR',
           refp,
           ite, forceTrainValSplit, outputSel))
    timepth = Path(resfldr,
       'Reference_RA_{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'MASknee_Times',
           refp,
           ite, forceTrainValSplit, outputSel))
    
    
    tsX = sX[:,inds]
    tvX = vX[:,inds]
    (tsXi,tsXj) = np.shape(tsX)
    (tvXi,tvXj) = np.shape(tvX)
    
    
    (pt0,wt0) = _timestamp() # Train model
    
    W, R = emlm.train(tsX,nY,beta=np.sqrt(np.finfo(float).eps),
              K_rel=refp,rs_method="RS-maximin", method='svd', weights=None)
    
    (pt1,wt1) = _timestamp()
    # calculate MAS
    rvals = MasFSreg(tsX, R, W, printT=False)
    
    (pt2,wt2) = _timestamp()
    
    # Kneepoint
    if len(rvals)==1:
        rankordr = np.argsort(rvals[0])
        rvals = rvals[0]
    else:
        rankordr = np.argsort(rvals)
    
    orderedRVals = np.flip(rvals[rankordr])
    # orderedRVals is now best to worst order
    orderedRValsX = np.arange(len(orderedRVals))
    
    (pt3,wt3) = _timestamp()
    
    # Kneepoint
    kneedle = KneeLocator(orderedRValsX,orderedRVals,curve='convex',direction='decreasing',
                          interp_method="interp1d", online=True)
    
    (pt4,wt4) = _timestamp()
    
    kneeval = kneedle.knee_y 
    # Figure out the true index of the kneepoint
    kneeInd = np.where(kneeval == rvals)[0][0]
    kneeIndOrdr = np.where(kneeInd == rankordr)[0][0]
    
    toBeRemovedInds = rankordr[:kneeIndOrdr]

    Nf = len(rankordr)
    
    toBeKeptBool = np.ones(Nf,dtype=bool)
    toBeKeptBool[toBeRemovedInds] = False

    tind = deepcopy(inds)
    

    tt = tind[tind]

    tind[tind] = np.logical_and(tt, toBeKeptBool)

    featureInds = np.logical_and(
            inds,tind)
    
    tsX = sX[:,featureInds]
    tvX = vX[:,featureInds]
    
    (pt5,wt5) = _timestamp()
    
    W, R = emlm.train(tsX,nY,beta=np.sqrt(np.finfo(float).eps),
              K_rel=refp,rs_method="RS-maximin", method='svd', weights=None)
    
    (pt6,wt6) = _timestamp()
    
    predY = emlm.pred2(tvX, W, R)
    
    (pt7,wt7) = _timestamp()
    
    valErr = RMSE(vY,predY.reshape((-1,1)))
    
    (pt8,wt8) = _timestamp()
    
    """
      train   MAS     ordr    knecre  stuff   train   pred    rmse
    [[pt1-pt0,pt2-pt1,pt3-pt2,pt4-pt3,pt5-pt4,pt6-pt5,pt7-pt6,pt8-pt7,
      wt1-wt0,wt2-wt1,wt3-wt2,wt4-wt3,wt5-wt4,wt6-wt5,wt7-wt6,wt8-wt7]]
    """

    times = [pt1-pt0,pt2-pt1,pt3-pt2,pt4-pt3,pt5-pt4,pt6-pt5,pt7-pt6,pt8-pt7,
             wt1-wt0,wt2-wt1,wt3-wt2,wt4-wt3,wt5-wt4,wt6-wt5,wt7-wt6,wt8-wt7]
    
    np.save(errpth,valErr)
    np.save(rankpth,rvals)
    np.save(timepth,times)


def FS_MASknee_N(N, inds, sX, nY, vX, vY, sYsclr, 
                 forceTrainValSplit=False,
                 regression=True, savekey='', savenum=-1,
                 savefldr='',refP=85, resfldr='',
                 datakey='',outputSel=-1,ite=0,
                 namePrefix=None):
    """
    A single function version of the MASvec-kneepoint feature selection 
    algorithm. Repeats the process given N times.
    """
    print('FS_MASknee')
    
    print('refp:',refP)
    
    if namePrefix is None:
        nmPrfx = 'Reference_RA_'
    else:
        nmPrfx = ''
    
    featpth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'MASkneeN{0:d}_Feats'.format(N),
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    errpth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'MASkneeN{0:d}_Err'.format(N),
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    rankpth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'MASkneeN{0:d}_featR'.format(N),
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    timepth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'MASkneeN{0:d}_Times'.format(N),
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    
    # Savespace initializations
    
    
    
    
    featureInds = deepcopy(inds)
    fIndsLog = deepcopy(inds)
    timestamps = [_timestamp(-1)]
    
    for iteration in range(N):
        
        if np.sum(featureInds) <= 1:
            timestamps.append(_timestamp(10))
            break
        
        tsX = sX[:,featureInds]
        tvX = vX[:,featureInds]
        (tsXi,tsXj) = np.shape(tsX)
        (tvXi,tvXj) = np.shape(tvX)
        
        timestamps.append(_timestamp(0))
        
        W, R = emlm.train(tsX,nY,beta=np.sqrt(np.finfo(float).eps),
                  K_rel=refP,rs_method="RS-maximin", method='svd', weights=None)
        
        timestamps.append(_timestamp(1))# calculate MAS
        
        rvals = MasFSreg(tsX, R, W, printT=False)
        
        print('shape(rvals)',np.shape(rvals))
        print(rvals)
        
        timestamps.append(_timestamp(2))
        
        # Kneepoint
        if len(rvals)==1:
            rankordr = np.argsort(rvals[0])
            rvals = rvals[0]
        else:
            rankordr = np.argsort(rvals)
        
        orderedRVals = np.flip(rvals[rankordr])
        # orderedRVals is now best to worst order
        orderedRValsX = np.arange(len(orderedRVals))
        
        timestamps.append(_timestamp(3))
        
        # Kneepoint
        kneedle = KneeLocator(orderedRValsX,orderedRVals,curve='convex',direction='decreasing',
                              interp_method="interp1d", online=True)
        
        timestamps.append(_timestamp(4))
        
        kneeval = kneedle.knee_y 
        if kneeval is None:
            # When there is no kneepoint
            timestamps.append(_timestamp(11))
            break
        
        # Figure out the true index of the kneepoint
        kneeInd = np.where(kneeval == rvals)[0][0]
        kneeIndOrdr = np.where(kneeInd == rankordr)[0][0]
        
        toBeRemovedInds = rankordr[:kneeIndOrdr]
    
        Nf = len(rankordr)
        
        toBeKeptBool = np.ones(Nf,dtype=bool)
        toBeKeptBool[toBeRemovedInds] = False
    
        tind = deepcopy(featureInds)
        
    
        tt = tind[tind]
    
        tind[tind] = np.logical_and(tt, toBeKeptBool)
    
        featureInds = np.logical_and(
                featureInds,tind)
        fIndsLog = np.vstack((fIndsLog,featureInds))
        
        timestamps.append(_timestamp(5))
    
    tsX = sX[:,featureInds]
    tvX = vX[:,featureInds]
    
    timestamps.append(_timestamp(6))
    
    W, R = emlm.train(tsX,nY,beta=np.sqrt(np.finfo(float).eps),
              K_rel=refP,rs_method="RS-maximin", method='svd', weights=None)
    
    timestamps.append(_timestamp(7))
    
    predY = emlm.pred2(tvX, W, R)
    
    timestamps.append(_timestamp(8))
    
    valErr = RMSE(vY,predY.reshape((-1,1)))
    
    timestamps.append(_timestamp(9))
    
    """
      train   MAS     ordr    knecre  stuff   train   pred    rmse
    [[pt1-pt0,pt2-pt1,pt3-pt2,pt4-pt3,pt5-pt4,pt6-pt5,pt7-pt6,pt8-pt7,
      wt1-wt0,wt2-wt1,wt3-wt2,wt4-wt3,wt5-wt4,wt6-wt5,wt7-wt6,wt8-wt7]]
    """
    
    
    np.save(featpth,fIndsLog)
    np.save(errpth,valErr)
    np.save(rankpth,rvals)
    np.save(timepth,timestamps)
    
    
def _wrapper_ComputeErrors(refP,tX,vX,nY,vY,inds):
    """
    """
    tsX = tX[:,inds]
    tvX = vX[:,inds]
    
    (_, _, numI) = rps(tsX, nY, 66, 'RS-maximin', returnI=True, directK=False,
        forceCalculation=True)
    N = len(tsX)
    
    #Inds = np.arange(N)
    cI = np.ones(N,dtype=np.int)
    cI[numI] = 0
    I = np.zeros(N,dtype=np.int)
    I[numI] = 1
    
    
    trainX = tsX[I,:]
    trainY = nY[I,:]
    testX = tsX[cI,:]
    testY = nY[cI,:]
    
    W, R = emlm.train(trainX,trainY,beta=np.sqrt(np.finfo(float).eps),
              K_rel=refP,rs_method="RS-maximin", method='svd', weights=None)
    
    predTrnY = emlm.pred2(trainX, W, R)
    trnErr = RMSE(trainY,predTrnY.reshape((-1,1)))
    
    predTstY = emlm.pred2(testX, W, R)
    tstErr = RMSE(testY,predTstY.reshape((-1,1)))
    
    
    predValY = emlm.pred2(tvX, W, R)
    valErr = RMSE(vY,predValY.reshape((-1,1)))
    
    return (trnErr,tstErr,valErr)


def FS_Wrapper(inds, sX, nY, vX, vY, sYsclr, 
                 forceTrainValSplit=False,
                 regression=True, savekey='', savenum=-1,
                 savefldr='',refP=85, resfldr='',
                 datakey='',outputSel=-1,ite=0,
                 wrapperImprFailN=10,
                 namePrefix=None,
                 autoStop=True):
    """
    """
    
    if namePrefix is None:
        nmPrfx = 'Reference_RA_'
    else:
        nmPrfx = ''
    
    
    featpth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'Wrapper_Feats',
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    errpth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'Wrapper_Err',
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    rankpth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'Wrapper_featR',
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    timepth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'Wrapper_Times',
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    
    timestamps = [_timestamp(-1)]

    if savekey is None:
        raise NotImplementedError
    else:
        #isregression = regression
        #svkey = savekey
        svnum = savenum

    improvementLimit = True
    if wrapperImprFailN==-1:
        improvementLimit = False
        
        
    featureInds = deepcopy(inds)
    featureInds = featureInds.reshape((1,-1))
    print(featureInds, np.shape(featureInds))
    
    fIndsLog = deepcopy(inds)
    tsX = sX[:,inds]
    tvX = vX[:,inds]
    (tsXi,tsXj) = np.shape(tsX)
    (tvXi,tvXj) = np.shape(tvX)
    
    print(np.shape(tsX))
    
    timestamps.append(_timestamp(0))# Train model
    
    W, R = emlm.train(tsX,nY,beta=np.sqrt(np.finfo(float).eps),
              K_rel=refP,rs_method="RS-maximin", method='svd', weights=None)
    
    timestamps.append(_timestamp(1))# calculate MAS
    
    rvals = MasFSreg(tsX, R, W, printT=False)   
    print('rvals',rvals, np.shape(rvals))
    
    timestamps.append(_timestamp(2))
    
    errors = _wrapper_ComputeErrors(refP,sX,vX,nY,vY,featureInds[-1,:])
    (trnErr,tstErr,valErr) = errors
    
    featureErrors = errors
    
    timestamps.append(_timestamp(3))
    
    if len(rvals)==1:
        rankordr = np.argsort(rvals[0])
        #featRanking = rvals[0]
    else:
        rankordr = np.argsort(rvals)
        #featRanking = rvals
        
    print('rankordr',rankordr,np.shape(rankordr))

    Nf = len(rankordr)
    print('Length of rankvec:',Nf)
    BestTrn = trnErr
    ref_errFs = tstErr
    improvementCounter = 0

    failFlag = False
    
    print('featureInds[-1,:]', np.shape(featureInds[-1,:]),np.sum(featureInds[-1,:]))
    
    CNV = (np.arange(np.shape(featureInds)[1]))[featureInds[-1,:]]
    
    bestK = 0
    for k in range(1,len(rankordr)):
        timestamps.append(_timestamp(4))
        
        FI = deepcopy(featureInds[-1,:])
        
        
        TBR = rankordr[:k] # to be removed inds in rankordr
        FI[CNV[TBR]] = False
        
        nAV = np.sum(FI)

        featureInds = np.vstack((featureInds, FI))
        fIndsLog = np.vstack((fIndsLog,featureInds[-1,:]))
        
        print('featureInds[-1,:]',np.shape(featureInds[-1,:]),np.sum(featureInds[-1,:]))
        print(np.shape(tsX))
        
        timestamps.append(_timestamp(5))
        
        errors = _wrapper_ComputeErrors(refP,sX,vX,nY,vY,featureInds[-1,:])
        
        timestamps.append(_timestamp(6))

        svnum+=1

        featureErrors = np.vstack((featureErrors, errors))

        # nTrn, nTst, nVal, oTrn, oTst, oVal
        (trnErr,tstErr,valErr) = errors


        if tstErr < ref_errFs:
            ref_errFs = tstErr
            BestTrn = trnErr
            nBest = nAV
            bestK = k
            timestamps.append(_timestamp(7))
            #W = WFs
            #XI = XIFs
            #SelVars = J;
            print('Rank (k = {0:1d}): n = {1:3d},\n  TrnFs = {2:1.3e}, TstFS = {3:1.3e}'.format(
                k, nBest, BestTrn, ref_errFs))
            improvementCounter = 0
        else:
            print('Rank (k = {0:1d}): n = {1:3d},\n  TrnFs = {2:1.3e}, TstFS = {3:1.3e} (NO IMPROVEMENT)'.format(
                k, nAV, trnErr, tstErr))
            improvementCounter += 1
            if (improvementCounter >= wrapperImprFailN and improvementLimit):
                if not failFlag:
                    print('Consecutive improvement failure limit reached.')
                    #failIter = k
                    failFlag = True
                if autoStop:
                    break
    
    if autoStop:
        print('Best result was with k={0:d}'.format(bestK))
        bki = bestK+1
        featureInds = featureInds[:bki,:]
        featureErrors = featureErrors[:bki,:]
                    
                    
    timestamps.append(_timestamp(7))
                    
    np.save(featpth, featureInds)
    np.save(errpth,  featureErrors)
    np.save(rankpth, rvals)
    np.save(timepth, timestamps)
    
    
def FS_Hybrid(inds, sX, nY, vX, vY, sYsclr, 
                 forceTrainValSplit=False,
                 regression=True, savekey='', savenum=-1,
                 savefldr='',refP=85, resfldr='',
                 datakey='',outputSel=-1,ite=0,
                 wrapperImprFailN=10,
                 namePrefix=None,
                 autoStop=True):
    """
    """
    
    if namePrefix is None:
        nmPrfx = 'Reference_RA_'
    else:
        nmPrfx = ''
    
    featpth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'Hybrid_Feats',
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    errpth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'Hybrid_Err',
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    rankpth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'Hybrid_featR',
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    timepth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'Hybrid_Times',
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    
    timestamps = [_timestamp(-1)]

    if savekey is None:
        raise NotImplementedError
    else:
        #isregression = regression
        #svkey = savekey
        svnum = savenum

    improvementLimit = True
    if wrapperImprFailN==-1:
        improvementLimit = False
        
        
    featureInds = deepcopy(inds)
    featureInds = featureInds.reshape((1,-1))
    
    
    print('featureInds[-1,:]', np.shape(featureInds[-1,:]),np.sum(featureInds[-1,:]))
    
    tsX = sX[:,featureInds[-1,:]]
    tvX = vX[:,featureInds[-1,:]]
    (tsXi,tsXj) = np.shape(tsX)
    (tvXi,tvXj) = np.shape(tvX)
    
    timestamps.append(_timestamp(1))
    
    W, R = emlm.train(tsX,nY,beta=np.sqrt(np.finfo(float).eps),
              K_rel=refP,rs_method="RS-maximin", method='svd', weights=None)
    
    timestamps.append(_timestamp(2))# calculate MAS
    
    rvals = MasFSreg(tsX, R, W, printT=False)
    
    print('shape(rvals)',np.shape(rvals))
    
    timestamps.append(_timestamp(2))
    
    # Kneepoint
    if len(rvals)==1:
        rankordr = np.argsort(rvals[0])
        rvals = rvals[0]
    else:
        rankordr = np.argsort(rvals)
        
    print('Rankordr',rankordr,np.shape(rankordr))
    
    orderedRVals = np.flip(rvals[rankordr])
    # orderedRVals is now best to worst order
    orderedRValsX = np.arange(len(orderedRVals))
    
    timestamps.append(_timestamp(3))
    
    # Kneepoint
    kneedle = KneeLocator(orderedRValsX,orderedRVals,curve='convex',direction='decreasing',
                          interp_method="interp1d", online=True)
    
    timestamps.append(_timestamp(4))
    
    kneeval = kneedle.knee_y 
    # Figure out the true index of the kneepoint
    kneeInd = np.where(kneeval == rvals)[0][0]
    kneeIndOrdr = np.where(kneeInd == rankordr)[0][0]
    
    toBeRemovedInds = rankordr[:kneeIndOrdr]

    Nf = len(rankordr)
    
    toBeKeptBool = np.ones(Nf,dtype=bool)
    toBeKeptBool[toBeRemovedInds] = False

    tind = deepcopy(featureInds[-1,:])
    

    tt = tind[tind]

    tind[tind] = np.logical_and(tt, toBeKeptBool)
    

    featureInds = np.vstack((featureInds,
             np.logical_and(featureInds[-1,:], tind)))
    
    
    print('featureInds[-1,:]', np.shape(featureInds[-1,:]),np.sum(featureInds[-1,:]))
    
    timestamps.append(_timestamp(5))
    
    tsX = sX[:,featureInds[-1,:]]
    tvX = vX[:,featureInds[-1,:]]
    
    W, R = emlm.train(tsX,nY,beta=np.sqrt(np.finfo(float).eps),
              K_rel=refP,rs_method="RS-maximin", method='svd', weights=None)
    
    rvals = MasFSreg(tsX, R, W, printT=False)   
    
    if len(rvals)==1:
        rankordr = np.argsort(rvals[0])
        rvals = rvals[0]
    else:
        rankordr = np.argsort(rvals)
    
    errors = _wrapper_ComputeErrors(refP,sX,vX,nY,vY,featureInds[-1,:])
    (trnErr,tstErr,valErr) = errors
    
    featureErrors = errors
    
    timestamps.append(_timestamp(6))

    Nf = len(rankordr)
    print('Length of rankvec:',Nf)
    BestTrn = trnErr
    ref_errFs = tstErr
    improvementCounter = 0
    
    
    failFlag = False
    
    print('featureInds[-1,:]', np.shape(featureInds[-1,:]),np.sum(featureInds[-1,:]))
    
    CNV = (np.arange(np.shape(featureInds)[1]))[featureInds[-1,:]]
    
    print('CNV',CNV,np.shape(CNV))
    print('Loop start')
    bestK = 0
    for k in range(1,len(rankordr)):
        timestamps.append(_timestamp(7))
        
        FI = deepcopy(featureInds[-1,:])
        
        
        TBR = rankordr[:k] # to be removed inds in rankordr
        print('TBR',TBR)
        FI[CNV[TBR]] = False
        
        nAV = np.sum(FI)

        featureInds = np.vstack((featureInds, FI))
        
        print('featureInds[-1,:]',np.shape(featureInds[-1,:]),np.sum(featureInds[-1,:]))
        print(np.shape(tsX))
        
        timestamps.append(_timestamp(8))
        
        errors = _wrapper_ComputeErrors(refP,sX,vX,nY,vY,featureInds[-1,:])
        
        timestamps.append(_timestamp(9))

        svnum+=1

        featureErrors = np.vstack((featureErrors, errors))

        # nTrn, nTst, nVal, oTrn, oTst, oVal
        (trnErr,tstErr,valErr) = errors


        if tstErr < ref_errFs:
            ref_errFs = tstErr
            BestTrn = trnErr
            nBest = nAV
            bestK = k
            timestamps.append(_timestamp(10))
            #W = WFs
            #XI = XIFs
            #SelVars = J;
            print('Rank (k = {0:1d}): n = {1:3d},\n  TrnFs = {2:1.3e}, TstFS = {3:1.3e}'.format(
                k, nBest, BestTrn, ref_errFs))
            improvementCounter = 0
        else:
            print('Rank (k = {0:1d}): n = {1:3d},\n  TrnFs = {2:1.3e}, TstFS = {3:1.3e} (NO IMPROVEMENT)'.format(
                k, nAV, trnErr, tstErr))
            improvementCounter += 1
            if (improvementCounter >= wrapperImprFailN and improvementLimit):
                if not failFlag:
                    print('Consecutive improvement failure limit reached.')
                    #failIter = k
                    failFlag = True
                if autoStop:
                    break
    
    if autoStop:
        print('Best result was with k={0:d}'.format(bestK))
        bki = bestK+1
        featureInds = featureInds[:bki,:]
        featureErrors = featureErrors[:bki,:]
                    
                    
    timestamps.append(_timestamp(11))
                    
    np.save(featpth, featureInds)
    np.save(errpth,  featureErrors)
    np.save(rankpth, rvals)
    np.save(timepth, timestamps)
    
    

def PureEMLM(inds, sX, nY, vX, vY, sYsclr, 
                 forceTrainValSplit=False,
                 regression=True, savekey='', savenum=-1,
                 savefldr='',refP=85, resfldr='',
                 datakey='',outputSel=-1,ite=0,
                 wrapperImprFailN=10,
                 namePrefix=None,):
    """
    """
    
    if namePrefix is None:
        nmPrfx = 'Reference_RA_'
    else:
        nmPrfx = ''
    
    featpth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'EMLM_Feats',
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    errpth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'EMLM_Err',
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    timepth = Path(resfldr,
       '{6:s}{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
           datakey,'EMLM_Times',
           refP,
           ite, forceTrainValSplit, outputSel, nmPrfx))
    
    timestamps = [_timestamp(-1)]
    
    
    tsX = sX[:,inds]
    tvX = vX[:,inds]
    (tsXi,tsXj) = np.shape(tsX)
    (tvXi,tvXj) = np.shape(tvX)
    
    
    timestamps.append(_timestamp(0))
    
    W, R = emlm.train(tsX,nY,beta=np.sqrt(np.finfo(float).eps),
              K_rel=refP,rs_method="RS-maximin", method='svd', weights=None)
    
    timestamps.append(_timestamp(1))
    
    predY = emlm.pred2(tvX, W, R)
    
    timestamps.append(_timestamp(2))
    
    valErr = RMSE(vY,predY.reshape((-1,1)))
    
    timestamps.append(_timestamp(3))
    
    print('Validation error: ',valErr)
    
    np.save(featpth,inds)
    np.save(errpth,valErr)
    np.save(timepth,timestamps)
    
    
