#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This source code was used in article 
"Distance-based feature selection for physical data" by Linja et al. 
Published in Neurocomputation

This script was made to study the feature ranking methods themselves using 
synthetic datasets. The correct rank of features is known with the synthetic 
datasets and this allows for ranking methods to be studied. 

EMLM related codes based on Matlab codes by Tommi Kärkkäinen and Joonas Hämäläinen,
@author: Joakim Linja
"""

import numpy as np
from pathlib import Path
from matlabLike import pdist2
from datasets import getDatasets
from scaling import MinMax
from ML_functions import (
                          RMSE,
                          solve_Axb,
                          preprocess,
                          )
from featureSelections import MasFSreg
import emlm
from mlm import rps
from copy import deepcopy
import os
from kneed import KneeLocator
import referenceFSmethods as refFSmethods
from time import time, process_time
from featureSelections import FS_MASknee

_datasetKeys = {
            'Artificial':1,
            'C00-2_1000':1,
            'C01-2_1000':1,
            'C02-2_1000':1,
            'C03-2_2000':1,
            'C04-10_2000':1,
            'C05-5_2000':1,
            'C06-6_2000':1,
            'C07-6_2000':1,
            'C08-6_3000':0,
            'C09-8_3000':0,
            'C08-6_3000':0,
            'C09-8_3000':0,
            'C10-15_3000':1,
            'C11-10_5000':1,
            'C12-10_5000':1,
            'C13-10_3000':1,
            'C14-10_3000':1,
            'StudentTest':0,
            'ATP1D':1,
            'Ionosphere':0,
            'COIL':0,
            'Madelon':0,
            'Outdoor':0,
            'ThyroidAnn':0,
            'OptDigits':0,
            'SatImage':0,
            'COIL2000':0,
            'Gisette':0,
            'PenDigits':0,
            'RF2':1,
            'ComputerActivity':1,
            'SCM1D':1,
            'IndoorLoc':1,
            'Census':1,
            'HNPAI':1,
            'Mnist':0,
            'Au38Q-N100-12k':1,
            'Au38Q-N100-8k':1,
            'Au38Q-N100-4k':1,
            'Au38Q-N10-12k':1,
            'Au38Q-N10-8k':1,
            'Au38Q-N10-4k':1,
            'Au38Q-N2-12k':1,
            'Au38Q-N2-8k':1,
            'Au38Q-N2-4k':1,
            'HNPAI-orig':1,
            'HNPAI-orig2':1,
            'The100_1000':1,
            'synth_A1_N1000_p100_ns1':1,
            'synth_A1_N1000_p200_ns1':1,
            'synth_A1_N1000_p2000_ns1':1,
            'synth_A2_N1000_p100_ns1':1,
            'synth_A2_N1000_p200_ns1':1,
            'synth_A2_N1000_p2000_ns1':1,
            'synth_A3_N1000_p100_ns1':1,
            'synth_A3_N1000_p200_ns1':1,
            'synth_A3_N1000_p2000_ns1':1,
            'synth_A4_N1000_p100_ns1':1,
            'synth_A4_N1000_p200_ns1':1,
            'synth_A4_N1000_p2000_ns1':1,
            'synth_A5_N1000_p100_ns1':1,
            'synth_A5_N1000_p200_ns1':1,
            'synth_A5_N1000_p2000_ns1':1,
            'synth_A6_N1000_p100_ns1':1,
            'synth_A6_N1000_p200_ns1':1,
            'synth_A6_N1000_p2000_ns1':1,
            'synth_R1_N1000_p100_ns1':1,
            'synth_R1_N1000_p200_ns1':1,
            'synth_R1_N1000_p2000_ns1':1,
            'synth_R2_N1000_p100_ns1':1,
            'synth_R2_N1000_p200_ns1':1,
            'synth_R2_N1000_p2000_ns1':1,
            'synth_R3_N1000_p100_ns1':1,
            'synth_R3_N1000_p200_ns1':1,
            'synth_R3_N1000_p2000_ns1':1,
            'synth_R4_N1000_p100_ns1':1,
            'synth_R4_N1000_p200_ns1':1,
            'synth_R4_N1000_p2000_ns1':1,
            'synth_A1_N1000_p100_ns0':1,
            'synth_A1_N1000_p200_ns0':1,
            'synth_A1_N1000_p2000_ns0':1,
            'synth_A2_N1000_p100_ns0':1,
            'synth_A2_N1000_p200_ns0':1,
            'synth_A2_N1000_p2000_ns0':1,
            'synth_A3_N1000_p100_ns0':1,
            'synth_A3_N1000_p200_ns0':1,
            'synth_A3_N1000_p2000_ns0':1,
            'synth_A4_N1000_p100_ns0':1,
            'synth_A4_N1000_p200_ns0':1,
            'synth_A4_N1000_p2000_ns0':1,
            'synth_A5_N1000_p100_ns0':1,
            'synth_A5_N1000_p200_ns0':1,
            'synth_A5_N1000_p2000_ns0':1,
            'synth_A6_N1000_p100_ns0':1,
            'synth_A6_N1000_p200_ns0':1,
            'synth_A6_N1000_p2000_ns0':1,
            'synth_R1_N1000_p100_ns0':1,
            'synth_R1_N1000_p200_ns0':1,
            'synth_R1_N1000_p2000_ns0':1,
            'synth_R2_N1000_p100_ns0':1,
            'synth_R2_N1000_p200_ns0':1,
            'synth_R2_N1000_p2000_ns0':1,
            'synth_R3_N1000_p100_ns0':1,
            'synth_R3_N1000_p200_ns0':1,
            'synth_R3_N1000_p2000_ns0':1,
            'synth_R4_N1000_p100_ns0':1,
            'synth_R4_N1000_p200_ns0':1,
            'synth_R4_N1000_p2000_ns0':1,
            'S1-lin':1,
            'S2-sin':1,
            'S3-pow2':1,
            'S4-pow6':1,
            'S5-exp':1,
            'S6-log':1,
            'R1-spe':1,
            'R2-ep':1,
            'R3-sin':1,
            'R4-sincos':1,
            }

_simulatedData = ['S1-lin', 'S2-sin', 'S3-pow2', 'S4-pow6', 'S5-exp',
                  'S6-log', 'R1-spe', 'R2-ep', 'R3-sin', 'R4-sincos']


def timestamp():
    return (process_time(), time())


def TrainEMLMregTs(X, Y, n1, beta, yscaler, fRefIndxs=None, fullPrint=False,
                   forceCalculation=False, weights=None):
    """

    Original doctext:

    function [TrnErr,TstErr,W,XI] = TrainEMLMregTs(X,Y,n1,beta)
    %
    %UNTITLED2 Determines right-inverse Extreme Minimal Learning Machine with
    % distance-based basis.
    %   (c) Tommi K., May 4, 2020 (ref. NeCo, 4/2019)
    % NOTE: This version computes testing error using that part of the whole
    % data that was not used as reference points.

    %
    % Selection of reference points for distance-based feature map
    %
    """

    # N = size(X,1);
    N = np.shape(X)[0]


    # I = rsmaxmin(X,n1);
    if fRefIndxs is None:
        (_, _, I) = rps(X, Y, n1, 'RS-maximin', returnI=True, directK=True,
            forceCalculation=forceCalculation)
        #(_, _, I) = rps(X, Y, n1,'random', returnI=True, directK=True,
        #    forceCalculation=forceCalculation)
    else:
        I = fRefIndxs[:n1]

    # XI = X(I,:); %(n1,n0)
    XI = X[I,:]
    (xi_n1, xi_n0) = np.shape(XI)
    assert xi_n1==n1
    # H = pdist2(XI,X); %(n1,N)
    H = pdist2(XI,X, weights=weights)
    (H_n1, H_N) = np.shape(H)
    assert H_n1 == n1
    assert H_N == N
    # H = pdist2(X,X)

    # W = (Y'*H') / (H*H' + (N*beta/n1)*eye(n1)); %(n2,n1)
    # rename: W = p1 / p2
    # In matlab this solves W*p2 = p1 for W
    # W @ p2 = p1 -> p2.T @ W.T = p1.T -> W = (np.linalg.solve(p2.T, p1.T)).T
    p1 = (Y.T) @ (H.T)
    p2 = H @ (H.T) + (N*beta/n1)*np.eye(n1)

    #W = (solve_Axb(p2.T, p1.T, method='svd')).T
    #W = (solve_Axb(p2.T, p1.T, method='lstsq')).T
    W = (solve_Axb(p2.T, p1.T, method='slstsq')).T


    #A/B = np.linalg.solve(B.conj().T, A.conj().T).conj().T
    #https://stackoverflow.com/questions/1007442/mrdivide-function-in-matlab-what-is-it-doing-and-how-can-i-do-it-in-python
    #import scipy as sp
    #W = (sp.linalg.lstsq(p2.conj().T, p1.conj().T)[0]).conj().T


    (W_n2, W_n1) = np.shape(W)
    assert W_n1 == n1

    # % Training error
    # Out = W*H; %NOTE: Scaling here if needed!
    Out = W @ H

    # %TrnErr = mean(sqrt(sum((Out'-Y).^2,2))); %MRSE, TommiK/SSPRWorkshop 2014
    # TrnErr = sqrt(mean((Out'-Y).^2)); %RMSE
    TrnErr = RMSE(Out.T, Y)

    oTrnErr = RMSE(yscaler.convertFromScale(Out.T),
                   yscaler.convertFromScale(Y))

    # % Testing error
    # J = setdiff(1:N,I);
    if len(I)==N:
        raise NotImplementedError('TrainEMLMregTs breaks if len(I)==N!')
    J = np.setdiff1d(np.arange(N),I)
    # H = H(:,J); %Distances for non RPs
    H = H[:,J]
    # Out = W*H; %NOTE: Scaling here if needed!
    Out = W @ H
    # %TstErr = mean(sqrt(sum((Out'-Y(J,:)).^2,2))); %MRSE, TommiK/SSPRWorkshop 2014
    # TstErr = sqrt(mean((Out'-Y(J)).^2)); %RMSE
    TstErr = RMSE(Out.T, Y[J])

    oTstErr = RMSE(yscaler.convertFromScale(Out.T),
                   yscaler.convertFromScale(Y[J]))

    return (TrnErr, TstErr, W, XI, oTrnErr, oTstErr, I, J)


def TrainEMLMregTs_Cont(X, Y, n1, beta, yscaler, savekey, savenum,
                        fRefIndxs=None, fullPrint=False, forceCalculation=False,
                        weights=None):
    """

    Original doctext:

    function [TrnErr,TstErr,W,XI] = TrainEMLMregTs(X,Y,n1,beta)
    %
    %UNTITLED2 Determines right-inverse Extreme Minimal Learning Machine with
    % distance-based basis.
    %   (c) Tommi K., May 4, 2020 (ref. NeCo, 4/2019)
    % NOTE: This version computes testing error using that part of the whole
    % data that was not used as reference points.

    %
    % Selection of reference points for distance-based feature map
    %
    """
    try:
        W = np.load('{0:s}_W_{1:d}.npy'.format(savekey, savenum))
        R = np.load('{0:s}_R_{1:d}.npy'.format(savekey, savenum))
        I = np.load('{0:s}_I_{1:d}.npy'.format(savekey, savenum))
    except FileNotFoundError:
        return TrainEMLMregTs(X, Y, n1, beta, yscaler,
            fRefIndxs=fRefIndxs, fullPrint=fullPrint,
            forceCalculation=forceCalculation, weights=weights), True

    N = np.shape(X)[0]

    # XI = X(I,:); %(n1,n0)
    #XI = R
    (xi_n1, xi_n0) = np.shape(R)
    assert xi_n1==n1

    H = pdist2(R,X, weights=weights)
    (H_n1, H_N) = np.shape(H)
    assert H_n1 == n1
    assert H_N == N


    (W_n2, W_n1) = np.shape(W)
    assert W_n1 == n1

    # % Training error
    Out = W @ H

    TrnErr = RMSE(Out.T, Y)

    oTrnErr = RMSE(yscaler.convertFromScale(Out.T),
                   yscaler.convertFromScale(Y))

    # % Testing error
    if len(I)==N:
        raise NotImplementedError('TrainEMLMregTs breaks if len(I)==N!')
    J = np.setdiff1d(np.arange(N),I)
    H = H[:,J]
    Out = W @ H

    TstErr = RMSE(Out.T, Y[J])

    oTstErr = RMSE(yscaler.convertFromScale(Out.T),
                   yscaler.convertFromScale(Y[J]))

    return ((TrnErr, TstErr, W, R, oTrnErr, oTstErr, I, J), False)


def kneePointDBFS(X,Y, Yscaler, featureInds, featureErrs, W, R, valX, valY, 
                  rankingAlg='MASvec',
                  ncW=None, ncR=None,
                  nonconstV=None,
                  tempSaveName=None,
                  regression=True, savekey=None, savenum=0,
                  savefldr=None,refP=85,
                  rankingSVpth=None,
                  timeArray=None):
    """
    """
    print('kneePointDBFS')

    if rankingSVpth is None:
        raise NotImplementedError

    if tempSaveName is None:
        useTempSave = False
    else:
        useTempSave = True

    if savekey is None:
        raise NotImplementedError
    else:
        isregression = regression
        svkey = savekey
        svnum = savenum
        
    if timeArray is None:
        raise NotImplementedError 

    fIselect = -1

    (xI,xJ) = np.shape(X)
    [trn_err, tst_err] = (featureErrs[fIselect,:])[slice(0,2)]
    XI = R
    
    if Yscaler is None:
        raise NotImplementedError('Yscaler missing')
        
    (pt0,wt0) = timestamp()
    rvals = MasFSreg(X[:,featureInds[fIselect,:]], XI, W, printT=False)
    (pt1,wt1) = timestamp()
    np.save(rankingSVpth, rvals)
    
    if len(rvals)==1:
        rankordr = np.argsort(rvals[0])
        rvals = rvals[0]
    else:
        rankordr = np.argsort(rvals)
    
    (pt2,wt2) = timestamp()
    orderedRVals = np.flip(rvals[rankordr])
    # orderedRVals is now best to worst order
    orderedRValsX = np.arange(len(orderedRVals))
    # Kneepoint
    kneedle = KneeLocator(orderedRValsX,orderedRVals,curve='convex',direction='decreasing',
                          interp_method="interp1d", online=True)
    (pt3,wt3) = timestamp()
    kneeval = kneedle.knee_y 
    # Figure out the true index of the kneepoint
    kneeInd = np.where(kneeval == rvals)[0][0]
    kneeIndOrdr = np.where(kneeInd == rankordr)[0][0]
    
    toBeRemovedInds = rankordr[:kneeIndOrdr]

    print('Length of rankvec:',len(rankordr))
    BestTrn = trn_err
    ref_errFs = tst_err

    fInd = np.shape(featureInds)[0] + fIselect

    Nf = len(rankordr)
    
    k = 1
    toBeKeptBool = np.ones(Nf,dtype=bool)
    
    toBeKeptBool[toBeRemovedInds] = False

    nAV = np.sum(toBeKeptBool)

    tind = deepcopy(featureInds[fInd,:])

    tt = tind[tind]

    tind[tind] = np.logical_and(tt, toBeKeptBool)

    featureInds = np.vstack((featureInds, np.logical_and(
            featureInds[fInd,:],tind)))
    
    (pt4,wt4) = timestamp()
    timeArr = np.vstack((timeArray,np.array([
        pt4-pt3,pt3-pt2,pt2-pt1,pt1-pt0,
        wt4-wt3,wt3-wt2,wt2-wt1,wt1-wt0])))

    (ters, WFs, XIFs, Is),times = computeErrors(
            featureInds[-1,:],X,Y,valX,valY,Yscaler,returnWR=True,
            returnNC=True, ncW=ncW, ncR=ncR, nonconstV=nonconstV,
            regression=isregression, savekey=svkey, savenum=svnum,
            savefldr=savefldr,refP=refP)
    svnum+=1
    print(times)
    timeArr = np.vstack((timeArr,np.array(times)))

    featureErrs = np.vstack((featureErrs, ters))

    # nTrn, nTst, nVal, oTrn, oTst, oVal
    (trn_errFs, tst_errFs, valErr) = ters[slice(0,3)]

    if useTempSave:
        np.save('featureInds_'+tempSaveName, featureInds)
        np.save('featureErrs_'+tempSaveName, featureErrs)


    failIter = -1
    if tst_errFs < ref_errFs:
        ref_errFs = tst_errFs
        BestTrn = trn_errFs
        nBest = nAV
        W = WFs
        XI = XIFs
        #SelVars = J;
        print('Rank (k = {0:1d}): n = {1:3d},\n  TrnFs = {2:1.3e}, TstFS = {3:1.3e}'.format(
            k, nBest, BestTrn, ref_errFs))
    else:
        print('Rank (k = {0:1d}): n = {1:3d},\n  TrnFs = {2:1.3e}, TstFS = {3:1.3e} (NO IMPROVEMENT)'.format(
            k, nAV, trn_errFs, tst_errFs))
    

    return (featureInds, featureErrs, failIter, timeArr)


def testSplitting(X, Y, xscaler, yscaler, outputSel=-1,
                 validationSet=None,
                 forceTrainValSplit=False,ite=0,
                 preprocessedTrainValSplit=True,
                 dkey=''):
    """
    Includes the ability to force own train/val split
    """
    print('testSplitting')

    oX = xscaler.convertFromScale(X)
    oY = yscaler.convertFromScale(Y)

    (N,xm) = np.shape(oX)
    valXconstFiltMask = None
    valYconstFiltMask = None

    if outputSel != -1:
        oY = oY[:,outputSel]
        oY = oY.reshape((N,1))

    if forceTrainValSplit:
        print('Forcing Train/Val split')


    if type(validationSet) is tuple and not forceTrainValSplit:
        print('Validation dataset readily provided')
        trnX = X
        (valX, valY) = validationSet
        valN = len(valX)
        if outputSel == -1:
            trnY = Y
        else:
            trnY = Y[:,outputSel].reshape((N,1))
            valY = valY[:,outputSel].reshape((valN,1))
        valXconstFiltMask = None
        valYconstFiltMask = None
        Xscaler = xscaler
        Yscaler = yscaler

    else:
        print('Taking validation dataset from training data with FoldSorter')
        if type(validationSet) is tuple and forceTrainValSplit:
            print('There is a readily provided validation dataset.')
            print('Combining it with the training data.')
            (valX, valY) = validationSet
            valN = len(valX)
            #print(valY)

            """
            validation set converter:
            (X, Y, nX, nY, Xscaler, Yscaler, XconstFiltMask, YconstFiltMask) = trainOut
            vX = valX[:,XconstFiltMask]
            vX = Xscaler.convertToScale(vX)
            vY = valY[:,YconstFiltMask]
            vY = Yscaler.convertToScale(vY)
            """

            ovalX = xscaler.convertFromScale(valX)
            ovalY = yscaler.convertFromScale(valY)

            # The training/validation sets combined:
            ocX = np.vstack((oX,ovalX))
            ocY = np.vstack((oY,ovalY))
        else:
            # If forcing is not done, rename
            ocX = oX
            ocY = oY

        # Separate ocX and ocY into nF separate folds
        print('Total N:',len(ocY))

        if preprocessedTrainValSplit:
            print('Utilizing preprocessed train/val split. Iteration:',ite)
            pptvsf = Path('Datasets/Saved_datasets/Splits')
            if len(dkey)==0:
                raise NotImplementedError('Required input variable "dkey" missing')
            trnInds = np.load(Path(pptvsf,'{0:s}_trnInds_o{1:d}_itrn{2:d}.npy'.format(
                dkey,outputSel,ite)))
            valInds = np.load(Path(pptvsf,'{0:s}_valInds_o{1:d}_itrn{2:d}.npy'.format(
                dkey,outputSel,ite)))

        else:
            raise NotImplementedError('This should not run, TrainValSplit should be preprocessed')

        ovalX = ocX[valInds,:]
        ovalY = ocY[valInds,:]

        trnX = ocX[trnInds,:]
        trnY = ocY[trnInds,:]

        trnXn0 = np.shape(trnX)

        (otrnX, otrnY, trnX, trnY, Xscaler, Yscaler,
             XconstFiltMask, YconstFiltMask) = \
                preprocess(trnX, trnY)

        trnXn1 = np.shape(trnX)
        if trnXn0[0]-trnXn1[0]>0:
            print('Reorganizing and repreprocessing removed: {0:d} observation(s)'.format(
                trnXn0[0]-trnXn1[0]))
        if trnXn0[1]-trnXn1[1]>0:
            print('Reorganizing and repreprocessing removed: {0:d} feature(s)'.format(
                trnXn0[1]-trnXn1[1]))

        valX = Xscaler.convertToScale(ovalX[:,XconstFiltMask])
        valY = Yscaler.convertToScale(ovalY[:,YconstFiltMask])
        valXconstFiltMask = XconstFiltMask
        valYconstFiltMask = YconstFiltMask

    ntsX = None
    ntsY = None
    print('Training/Validation split: {0:d}/{1:d} (total: {2:d})'.format(
        np.shape(trnX)[0],np.shape(valX)[0],np.shape(trnX)[0]+np.shape(valX)[0]))
    print('Unique outputs, trn:',len(np.unique(trnY)),'val:',len(np.unique(valY)))
    return (trnX, trnY, valX, valY, ntsX, ntsY,
            Xscaler, Yscaler, valXconstFiltMask, valYconstFiltMask)
    

def referenceFeatSel(inds,sX,nY,vX,valY,sYsclr,
              forceTrainValSplit=False,
              regression=True, savekey=None, savenum=-1,
              savefldr=None,refP=85, resfldr=None,
              datakey=None,outputSel=-1,ite=0):

    """
    """
    
    FS_MASknee(inds, sX, nY, vX, valY, sYsclr, 
                     forceTrainValSplit=forceTrainValSplit,
                     regression=regression, savekey=savekey, savenum=savenum,
                     savefldr=savefldr,refP=refP, resfldr=resfldr,
                     datakey=datakey,outputSel=outputSel,ite=ite)
    
    refFSmethods.ref_RandomForest(inds, sX, nY, vX, valY, sYsclr, 
                     forceTrainValSplit=forceTrainValSplit,
                     regression=regression, savekey=savekey, savenum=savenum,
                     savefldr=savefldr,refP=refP, resfldr=resfldr,
                     datakey=datakey,outputSel=outputSel,ite=ite)
    
    refFSmethods.ref_DecisionTree(inds, sX, nY, vX, valY, sYsclr, 
                     forceTrainValSplit=forceTrainValSplit,
                     regression=regression, savekey=savekey, savenum=savenum,
                     savefldr=savefldr,refP=refP, resfldr=resfldr,
                     datakey=datakey,outputSel=outputSel,ite=ite)
    
    refFSmethods.ref_LinearRegression(inds, sX, nY, vX, valY, sYsclr, 
                     forceTrainValSplit=forceTrainValSplit,
                     regression=regression, savekey=savekey, savenum=savenum,
                     savefldr=savefldr,refP=refP, resfldr=resfldr,
                     datakey=datakey,outputSel=outputSel,ite=ite)
    
    refFSmethods.ref_Lasso(inds, sX, nY, vX, valY, sYsclr, 
                     forceTrainValSplit=forceTrainValSplit,
                     regression=regression, savekey=savekey, savenum=savenum,
                     savefldr=savefldr,refP=refP, resfldr=resfldr,
                     datakey=datakey,outputSel=outputSel,ite=ite)
    
    refFSmethods.ref_RidgeRegression(inds, sX, nY, vX, valY, sYsclr, 
                     forceTrainValSplit=forceTrainValSplit,
                     regression=regression, savekey=savekey, savenum=savenum,
                     savefldr=savefldr,refP=refP, resfldr=resfldr,
                     datakey=datakey,outputSel=outputSel,ite=ite)
    
    refFSmethods.ref_SVM(inds, sX, nY, vX, valY, sYsclr, 
                     forceTrainValSplit=forceTrainValSplit,
                     regression=regression, savekey=savekey, savenum=savenum,
                     savefldr=savefldr,refP=refP, resfldr=resfldr,
                     datakey=datakey,outputSel=outputSel,ite=ite)


def computeErrors(inds, sX, sY, vX, vY, yscaler,
                  returnWR=False,
                  returnNC=False, ncW=None, ncR=None,
                  nonconstV=None,
                  regression=True, savekey=None, savenum=None, savefldr=None,
                  weights=None,RF=True,saveWRI=True,refP=85):
    """
    """
    #print('ComputeErrors')
    beta = np.sqrt(np.finfo(float).eps)
    (ij,) = np.shape(inds)
    (sXi,sXj) = np.shape(sX)
    (vXi,vXj) = np.shape(vX)
    (sYi,sYj) = np.shape(sY)
    (vYi,vYj) = np.shape(vY)
    try:
        assert ij == sXj
        assert sXj == vXj
        assert sYj == vYj
    except AssertionError as ae:
        print('Assertion Error')
        print('ij:',ij)
        print('sX:',sXi,sXj)
        print('vX:',vXi,vXj)
        print('sY:',sYi,sYj)
        print('vY:',vYi,vYj)
        raise AssertionError(ae)


    tsX = sX[:,inds]
    tvX = vX[:,inds]
    (tsXi,tsXj) = np.shape(tsX)
    (tvXi,tvXj) = np.shape(tvX)

    tw = weights
    try:
        assert tsXj == tvXj
    except AssertionError as ae:
        print('Assertion Error')
        print('tsX:',tsXi,tsXj)
        print('tvX:',tvXi,tvXj)
        raise AssertionError(ae)

    n1 = int(np.ceil((refP/100)*np.shape(sX)[0]))

    flg = False
    (pt0,wt0) = timestamp()
    if savekey is not None:
        if savenum is None:
            raise NotImplementedError
        ((trnErr, tstErr, W, R, oTrnErr, oTstErr, I, J), flg) = TrainEMLMregTs_Cont(
            tsX,sY,n1,beta,yscaler, savekey, savenum, weights=tw)
    else:
        (trnErr, tstErr, W, R, oTrnErr, oTstErr, I, J) = TrainEMLMregTs(
                tsX,sY,n1,beta,yscaler, weights=tw)
    (pt1,wt1) = timestamp()
    if savekey is not None and flg and saveWRI:
        if savenum is None:
            raise NotImplementedError
        if savefldr is None:
            np.save('{0:s}_W_{1:d}.npy'.format(savekey, savenum), W)
            np.save('{0:s}_R_{1:d}.npy'.format(savekey, savenum), R)
            np.save('{0:s}_I_{1:d}.npy'.format(savekey, savenum), I)
        else:
            np.save(Path(savefldr,'{0:s}_W_{1:d}.npy'.format(
                savekey, savenum)).as_posix(), W)
            np.save(Path(savefldr,'{0:s}_R_{1:d}.npy'.format(
                savekey, savenum)).as_posix(), R)
            np.save(Path(savefldr,'{0:s}_I_{1:d}.npy'.format(
                savekey, savenum)).as_posix(), I)

    (pt2,wt2) = timestamp()
    predY = emlm.pred2(tvX,W,R, weights=tw)
    (pt3,wt3) = timestamp()

    valErr1 = RMSE(vY,predY)
    (pt4,wt4) = timestamp()
    valErr2 = RMSE(yscaler.convertFromScale(vY),
                  yscaler.convertFromScale(predY))
    (pt5,wt5) = timestamp()

    if returnNC:
        # Compute validation error with the originally trained model
        if ncW is None or ncR is None:
            ncW = W
            ncR = R
        if nonconstV is None:
            raise NotImplementedError

        ncPredY = emlm.pred2(tvX, ncW, ncR[:, inds[nonconstV]])
        ncValErr1 = RMSE(vY, ncPredY)
        ncValErr2 = RMSE(yscaler.convertFromScale(vY),
                  yscaler.convertFromScale(ncPredY))


        if not regression:
            # Calculate misclassification errors
            testN = np.shape(vY)[0]
            oVY = yscaler.convertFromScale(vY)
            opY = yscaler.convertFromScale(predY)
            opY = np.round(opY).astype(np.int)
            misN = np.sum(oVY!=opY)
            misce = misN/testN
            misce2 = np.sum(np.abs(opY-oVY))/testN
        else:
            misce = -2
            misce2 = -2

    else:
        if not regression:
            # Calculate misclassification errors
            testN = np.shape(vY)[0]
            oVY = yscaler.convertFromScale(vY)
            opY = yscaler.convertFromScale(predY)
            misN = np.sum(oVY!=opY)
            misce = misN/testN
            misce2 = np.sum(np.abs(opY-oVY))/testN
        else:
            misce = -2
            misce2 = -2
        ncValErr1 = -2
        ncValErr2 = -2


    resAR = np.array([trnErr, tstErr, valErr1,
                  oTrnErr, oTstErr, valErr2,
                  ncValErr1, ncValErr2, misce, misce2])

    print('tstErr: {0:.3f}, valErr: {1:.3f}'.format(tstErr, valErr1))

    times = [pt5-pt4,pt4-pt3,pt3-pt2,pt1-pt0,
             wt5-wt4,wt4-wt3,wt3-wt2,wt1-wt0]

    if returnWR:
        return (resAR, W, R, I),times
    else:
        return resAR, times


def featureSelection(datS,outputSel=-1, fullPrint=False,
                     datakey=None,
                     validationSetP=0.1,
                     validationSet=None,
                     saveIntermRes=False,
                     rankingAlg='MASvec',
                     savefldr=None,refP=85,
                     forceTrainValSplit=False,
                     resfldr = None,
                     referenceResults=True,
                     ite=0,
                     preprocessedTrainValSplit=True,
                     rankingSVpth=None,
                     timepth=None):
    """
    """
    print('\nFeature selection:',datakey)
    
    if rankingSVpth is None:
        raise NotImplementedError

    oX, oY, nX, nY, xscaler, yscaler, XconstFiltMask, YconstFiltMask = datS

    isregression = 1 == _datasetKeys[datakey]
    svkey = 'RA_{0:s}_{1:s}'.format(datakey, rankingAlg)

    Nobs = len(nX)
    print('Dataset has',Nobs,'observations')
    if Nobs > 15000:
        useTempSaves = True
        tempSaveName = 'RA_{0:s}_{1:s}_temp.npy'.format(
                datakey,rankingAlg)
    else:
        useTempSaves = False
        tempSaveName = None


    (nX, nY, valX, valY, _, _,
     xscaler, yscaler, valXconstFiltMask, valYconstFiltMask) = testSplitting(
            nX, nY, xscaler, yscaler,           
            outputSel=outputSel, validationSet=validationSet,
            forceTrainValSplit=forceTrainValSplit, ite=ite,
            preprocessedTrainValSplit=preprocessedTrainValSplit,
            dkey=datakey)
         
    txc = XconstFiltMask

    if type(validationSet) is tuple and not forceTrainValSplit:
        print('Validation constant mask similarized')
        valXconstFiltMask = XconstFiltMask
    else:
        print('Updated constant mask with repreprocessing result')
        txc[XconstFiltMask] = np.logical_and(
            txc[XconstFiltMask], valXconstFiltMask)


    # Setup progress log variables
    # Actual oX cannot be used since it may be larger than nX due to
    # validation set splitting
    OX = xscaler.convertFromScale(nX)
    OY = yscaler.convertFromScale(nY)

    # sX and sY are for error computations and progress logging
    sXsclr = MinMax()
    sYsclr = MinMax()
    #sYsclr.Scaling(oY)
    sYsclr.Scaling(OY)

    # Initialize the progress logging variables. For Inds, True means feature
    # is still present at that point.
    featureInds = np.full((2,np.shape(oX)[1]),True,dtype=np.bool)
    #featureInds = np.full((2,np.shape(OX)[1]),True,dtype=np.bool)

    featureErrs = -1*np.ones((2,16)) # nTrn, nTst, nVal, oTrn, oTst, oVal
    # index [:,0] is the original where all features are present
    # index [:,1] is when constant features are removed.
    featureInds[-1,:] = np.logical_and(featureInds[-1,:],txc)

    print('Feature selection starts from {0:d} features'.format(
            np.sum(featureInds[0,:])))
    print('Non-constant features: {0:d} ({1:.1%})'.format(np.sum(featureInds[1,:]),
          np.sum(featureInds[1,:])/np.sum(featureInds[0,:])))

    # Prepare dataset s that can always be filtered with featureInds to produce
    # the current actual dataset.
    sX = sXsclr.Scaling(OX)
    tX = np.zeros((np.shape(sX)[0],np.shape(oX)[1]))
    vX = np.zeros((np.shape(valX)[0],np.shape(oX)[1]))

    tX[:,featureInds[-1,:]] = sX
    vX[:,featureInds[-1,:]] = valX


    constW = int(np.sum(np.logical_not(featureInds[-1,:])))
    tX[:,np.logical_not(featureInds[-1,:])] = -1*np.ones(
            (np.shape(sX)[0],constW))
    vX[:,np.logical_not(featureInds[-1,:])] = -1*np.ones(
            (np.shape(valX)[0],constW))
    sX = tX

    # Baseline calculation
    svnum = 0
    (featureErrs[-1,:], oW, oR, oI), times = \
        computeErrors(featureInds[-1,:],sX,nY,vX,valY,sYsclr,returnWR=True,
                      returnNC=True, nonconstV=XconstFiltMask,
                      regression=isregression, savekey=svkey, savenum=svnum,
                      savefldr=savefldr,refP=refP)
        
    if True:#referenceResults:
        referenceFeatSel(featureInds[-1,:],sX,nY,vX,valY,sYsclr,
                      forceTrainValSplit=forceTrainValSplit,
                      regression=isregression, savekey=svkey, savenum=svnum,
                      savefldr=savefldr,refP=refP, resfldr=resfldr,
                      datakey=datakey,outputSel=outputSel,ite=ite)
        
    print(times)
    timeArr = np.array(times)
    
    svnum +=1

    if useTempSaves:
        np.save('featureInds_'+tempSaveName, featureInds)
        np.save('featureErrs_'+tempSaveName, featureErrs)


    (ters,W,R,I),times = computeErrors(featureInds[-1,:],
                    sX, nY, vX, valY, sYsclr, returnWR=True,
                    returnNC=True, ncW=oW, ncR=oR, nonconstV=XconstFiltMask,
                    regression=isregression, savekey=svkey, savenum=svnum,
                    savefldr=savefldr,refP=refP)
    print(times)
    timeArr = np.vstack((timeArr,np.array(times)))
    svnum +=1
    featureErrs = np.vstack((featureErrs,ters))
    

    if useTempSaves:
        np.save('featureInds_'+tempSaveName, featureInds)
        np.save('featureErrs_'+tempSaveName, featureErrs)

    # Feature selection calculation
    (featureInds, featureErrs, failIter, timeArray) = kneePointDBFS(
        sX, nY, sYsclr, featureInds, featureErrs, W, R,
        vX, valY, 
        rankingAlg=rankingAlg,
        ncW=oW, ncR=oR,
        nonconstV=XconstFiltMask,
        tempSaveName=tempSaveName,
        regression=isregression, savekey=svkey, savenum=svnum,
        savefldr=savefldr,refP=refP,
        rankingSVpth=rankingSVpth, 
        timeArray=timeArr)
    
    if useTempSaves:
        fn1 = 'featureInds_'+tempSaveName
        fn2 = 'featureErrs_'+tempSaveName
        os.remove(fn1)
        os.remove(fn2)

    return (featureInds, featureErrs, failIter, timeArray)


def datasetFromFile(datakey,outpsel,ite,separateValidationSet=0):
    """
    """
    def helpfun(pth,add):
        return np.load(Path(pth,add+'.npy').as_posix())

    pth1 = Path('Datasets/Saved_datasets')
    pth2 = Path(pth1,'Params')
    X = helpfun(pth1,'{0:s}_origX'.format(datakey))
    Y = helpfun(pth1,'{0:s}_origY'.format(datakey))

    nX = helpfun(pth1,'{0:s}_normX'.format(datakey))
    nY = helpfun(pth1,'{0:s}_normY'.format(datakey))


    xcnstmsk = helpfun(pth2,'{0:s}_xconstmask'.format(datakey))
    ycnstmsk = helpfun(pth2,'{0:s}_yconstmask'.format(datakey))

    xmin = helpfun(pth2,'{0:s}_xmin'.format(datakey))
    xmax = helpfun(pth2,'{0:s}_xmax'.format(datakey))
    ymin = helpfun(pth2,'{0:s}_ymin'.format(datakey))
    ymax = helpfun(pth2,'{0:s}_ymax'.format(datakey))
    xscaler = MinMax()
    yscaler = MinMax()
    xscaler.setScalingParams((xmin,xmax))
    yscaler.setScalingParams((ymin,ymax))

    if separateValidationSet==1:
        print('Reading validation dataset')
        valX = helpfun(pth1,'{0:s}_valX'.format(datakey))
        valY = helpfun(pth1,'{0:s}_valY'.format(datakey))

        return ((X, Y, nX, nY, xscaler, yscaler, xcnstmsk, ycnstmsk),
                (valX, valY))

    (X, Y, nX, nY, xscaler, yscaler, xcnstmsk, ycnstmsk) = \
        preprocess(X, Y, unmodifiedOriginals=True)

    return ((X, Y, nX, nY, xscaler, yscaler, xcnstmsk, ycnstmsk),None)


def resprinter(fInds, fErrs):
    """
    """

    #np.array([trnErr, tstErr, valErr1,
    #              oTrnErr, oTstErr, valErr2,
    #              ncValErr1, ncValErr2, misce, misce2,
    #              rf_trn_err, rf_tst_err, rf_val_err,
    #              rf_otrn_err, rf_otst_err, rf_oval_err])

    fEi = 0
    bfErr = np.inf
    for i in range(2,len(fErrs)):
        t = fErrs[i,1]
        if t <= bfErr:
            bfErr = t
            fEi = i
    fI = fInds[fEi,:]
    print('Selected features:',np.sum(fI), 'Solution I:',fEi)



if __name__ == "__main__":
    print('')
    # General settings
    fP = False
    detDS = False

    readFromFile = True
    liveDataset = False#True

    unModOrig = True
    separateValidationSet = True
    validationSetP = 0.33
    saveIntermRes = False#True
    forceTrainValSplit = True
    #forceTrainValSplit = False

    rankingAlgs = [
                   'MASvec',
                   ]

    interFfldr = Path('WRI')
    if not interFfldr.exists():
        os.mkdir(interFfldr.as_posix())

    refRes = True

    refPs = [85]
    iterange = range(0,30)
    #iterange = [0]

    #"""
    dks = [
        #('synth_A1_N1000_p200_ns1',-1,0),
        #('synth_A2_N1000_p200_ns1',-1,0),
        #('synth_A3_N1000_p200_ns1',-1,0),
        #('synth_A4_N1000_p200_ns1',-1,0),
        #('synth_A5_N1000_p200_ns1',-1,0),
        #('synth_A6_N1000_p200_ns1',-1,0),
        #('synth_R1_N1000_p200_ns1',-1,0),
        #('synth_R2_N1000_p200_ns1',-1,0),
        #('synth_R3_N1000_p200_ns1',-1,0),
        #('synth_R4_N1000_p200_ns1',-1,0),
        ('ATP1D',0,0),                
        #('ATP1D',1,0),
        #('ATP1D',2,0),
        #('ATP1D',3,0),
        #('ATP1D',4,0),
        #('ATP1D',5,0),
        ('RF2',0,0),                  
        #('RF2',1,0),
        #('RF2',2,0),
        #('RF2',3,0),
        #('RF2',4,0),
        #('RF2',5,0),
        #('RF2',6,0),
        #('RF2',7,0),
        ('ComputerActivity',-1,0),
        ('SCM1D',0,0),                
        #('SCM1D',1,0),
        #('SCM1D',2,0),
        #('SCM1D',3,0),
        #('SCM1D',4,0),
        #('SCM1D',5,0),
        #('SCM1D',6,0),
        #('SCM1D',7,0),
        #('SCM1D',8,0),
        #('SCM1D',9,0),
        #('SCM1D',10,0),
        #('SCM1D',11,0),
        #('SCM1D',12,0),
        #('SCM1D',13,0),
        #('SCM1D',14,0),
        #('SCM1D',15,0),
        ('Census',-1,0),
        #('StudentTest',-1,1),         # (258,5/1)
        #('COIL',-1,1),                # (1800,20/1)#
        #('Madelon',-1,1),             # (2000,500/1)#
        #('Outdoor',-1,1),             # (2400,21/1)
        #('ThyroidAnn',-1,1),          # (3709,21/1)
        #('OptDigits',-1,1),           # (3823,62/1)
        #('SatImage',-1,1),            # (4435,36/1)
        #('COIL2000',-1,1),            # (5220,85/1)
        #('Gisette',-1,1),             # (6000,4955/1)
        #('PenDigits',-1,1),           # (7494,16/1)
        ]#"""

    datasetFs = getDatasets()
    for refP in refPs:#[1::2]:
        for rankingAlg in rankingAlgs:
                for dk in dks:
                    datakey = dk[0]
                    outputSel = dk[1]
                    print('Dataset:',datakey)
                    print('Output:',outputSel)
                    for ite in iterange:
                        print('Iter: ',ite)

                        savefolder = Path('Results/kneePointDBFS_vp{0:.2f}'.format(
                            validationSetP))
                        if not savefolder.exists():
                            os.mkdir(savefolder.as_posix())

                        pthInds = Path(savefolder,
                           'featureInds_RA_{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
                                   datakey,rankingAlg,
                                   refP,
                                   ite,forceTrainValSplit, outputSel))

                        pthErrs = Path(savefolder,
                           'featureErrs_RA_{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
                               datakey,rankingAlg,
                               refP,
                               ite,forceTrainValSplit, outputSel))
                        
                        pthFailIter = Path(savefolder,
                           'failIter_RA_{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
                               datakey,rankingAlg,
                               refP,
                               ite,forceTrainValSplit, outputSel))
                        
                        pthRANKvec = Path(savefolder,
                           'Ranking_RA_{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
                               datakey,rankingAlg,
                               refP,
                               ite,forceTrainValSplit, outputSel))
                        
                        pthTime = Path(savefolder,
                           'Time_RA_{0:s}_{1:s}_refP{2:03d}_fTV{4:d}_out{5:d}_i{3:d}.npy'.format(
                               datakey,rankingAlg,
                               refP,
                               ite,forceTrainValSplit, outputSel))

                        simS = datakey in _simulatedData

                        if readFromFile and not simS:
                            print('Reading preprocessed dataset from file')
                            (dataset,valset) = datasetFromFile(datakey,
                                            outputSel,ite,
                                            separateValidationSet=dk[2])
                        else:
                            raise NotImplementedError('Datasets should be preprocessed')

                        print('Reference points in RPS: {0:.2%}'.format(refP/100))

                        #"""
                        res = featureSelection(dataset, outputSel=outputSel,
                                               fullPrint=fP,
                                               datakey=datakey,
                                               validationSetP=validationSetP,
                                               validationSet=valset,
                                               saveIntermRes=saveIntermRes,
                                               rankingAlg=rankingAlg,
                                               savefldr=interFfldr,refP=refP,
                                               forceTrainValSplit=forceTrainValSplit,
                                               resfldr=savefolder,
                                               referenceResults=refRes,
                                               ite=ite,
                                               rankingSVpth=pthRANKvec,
                                               timepth=pthTime)
                    

                        (featureInds, featureErrs, failIter, timeArray) = res
                        

                        print(timeArray)
                        
                        np.save(pthInds.as_posix(),featureInds)
                        np.save(pthErrs.as_posix(),featureErrs)
                        np.save(pthFailIter.as_posix(),failIter)
                        np.save(pthTime.as_posix(),timeArray)

                        txtfrmt = '10s'
                        prntFrmt = '10f'
                        print('{6:4s} {0:{5:s}} {1:{5:s}} {2:{5:s}} {3:{5:s}} {4:{5:s}} {7:7s}'.format(
                                'nTrn', 'nTst', 'nVal',
                                'msc', 'msc2', txtfrmt,'i','p'))

                        for i in range(1,len(featureErrs)):
                            fe = featureErrs[i,:]
                            print('{6:2d} {0:{5:s}} {1:{5:s}} {2:{5:s}} {3:{5:s}} {4:{5:s}} {7:7d}'.format(
                                    fe[0],fe[1],fe[2],fe[8],fe[9],prntFrmt,i,np.sum(featureInds[i,:])))

                        print('Results for dataset saved at',time())
                        print('Location:')
                        print(pthInds)

                        resprinter(featureInds, featureErrs)

                        print('\n\n')
        #"""

    print('\nEnd of Line')