#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""


@author: Joonas Hämäläinen
@author: Joakim Linja

Modified by Joakim Linja:
    train-function now returns Dx and Dy instead of B so that
    different B-calculation methods can be tested
    pred-function input argument order has been changed to allow partial function
    
    The B-calculation functions are in file *Bfuns.py*
"""

#
# Minimal Learning Machine (MLM) implementation for 1D regression
#

import numpy as np
import warnings
import random
from pairwiseDistanceCalculation import pdc
from itertools import repeat
from pathlib import Path
from os import remove as osremove


# Train MLM
def train(X, Y, Bfun, K_rel=100, rs_method="random", printP=False):
    """
    Trains MLM model.

    Parameters
    ----------

    *X* : numpy.ndarray
        Input data matrix, rows are observations.

    *Y* : numpy.ndarray
        Output data matrix, rows are observations.
        
    *Bfun* : function
        Function which is used to solve the B-matrix in MLM. File *Bfuns.py* 
        contains premade ones. 

    *K_rel* : int, optional
        Relative #observations to be seleted as reference points.

        Default: 100

    *rs_method* : str, optional
        "random" uses random sampling to select reference points.

        "RS-maximin" uses RS-maximin to select reference points.

        Default: "random"

    *printP* : boolean, optional
        Sets whether the reference point selection method is printed to console. 
        
        Default: False


    Returns
    -------

    *Dx* : numpy.matrix
        Input space distance matrix

    *Dy* : numpy.matrix
        Output space distance matrix

    *R* : numpy.ndarray
        Input data reference points

    *T* : numpy.ndarray
        Output data reference points

    """

    # X and Y should be 2D ndarrays, e.g., shape should be (N,1)
    # not (N,)
    if len(X.shape) == 1:
        warnings.warn("Reshaping input data matrix...")
        X = X.reshape(-1,1)
    if len(Y.shape) == 1:
        warnings.warn("Reshaping output data matrix...")
        Y = Y.reshape(-1,1)

    # Check for constant variables
    X_max = np.max(X,axis=0)
    X_min = np.min(X,axis=0)
    is_const_var = X_max - X_min == 0

    if type(is_const_var)==np.ndarray:
        if any(is_const_var):
            warnings.warn("Input data matrix contains constant columns!" \
                      "Results may be inaccurate!")
    else:
        if any(list(map(lambda x: any(x), np.array(is_const_var)))):
            warnings.warn("Input data matrix contains constant columns!" \
                      "Results may be inaccurate!")

    if printP:
        print("Selecting reference points with " + rs_method + " method...")
    R,T = rps(X, Y, K_rel, rs_method)

    B, K = _train(X, Y, R, T, Bfun)

    return B, R, T


# Trains MLM with selected sets of reference points
def _train(X, Y, R, T, Bfun):
    """
    Trains MLM model.

    Parameters
    ----------

    *X* : numpy.ndarray
        Input data matrix, rows are observations.

    *Y* : numpy.ndarray
        Output data matrix, rows are observations.

    *R* : numpy.ndarray
        Input data reference points

    *T* : numpy.ndarray
        Output data reference points
        
    *Bfun* : function
        Function which is used to solve the B-matrix in MLM. File *Bfuns.py* 
        contains premade ones. 


    Returns
    -------

    *B* : numpy.ndarray
        Distance regression model coefficients.

    """
   
    if len(X.shape) == 1:
        warnings.warn("Reshaping input data matrix...")
        X = X.reshape(-1,1)
    if len(Y.shape) == 1:
        warnings.warn("Reshaping output data matrix...")
        Y = Y.reshape(-1,1)
    if len(R.shape) == 1:
        warnings.warn("Reshaping input data reference point matrix...")
        R = R.reshape(-1,1)
    if len(T.shape) == 1:
        warnings.warn("Reshaping output data reference point matrix...")
        T = T.reshape(-1,1)

    # Compute the distance matrices
    #Dx = euclidean_distances(X,R) # Original
    #Dy = euclidean_distances(Y,T) # Original
    Dx = pdc(X,R)
    Dy = pdc(Y,T)

    # Compute B = (Dx^T Dx)^-1 (Dx^T Dy)
    K = len(T)

    # Original left as a comment: 
    #DxDx = np.matmul(Dx.transpose(),Dx) + np.sqrt(np.finfo(float).eps)*np.eye(K)
    #DxDy = np.matmul(Dx.transpose(),Dy)

    #B = np.linalg.lstsq(np.matmul(Dx.transpose(),Dx) +\
    #                    np.sqrt(np.finfo(float).eps)*np.eye(K),\
    #                    np.matmul(Dx.transpose(),Dy))[0]
    return Bfun(Dx,Dy), K


# Reference Points Selection for MLM
def rps(X, Y, K_rel, rs_method, returnI=False, directK=False,
        forceCalculation=False, progressingI=False,
        saveInterval=10000):
    """
    Selects reference points for the MLM model.

    Parameters
    ----------

    *X* : numpy.ndarray
        Input data matrix, rows are observations.

    *Y* : numpy.ndarray
        Output data matrix, rows are observations.

    *K_rel* : int
        Relative #observations to be seleted as reference points.

    *rs_method* : str
        "random" uses random sampling to select reference points. 
        
        "RS-maximin" uses RS-maximin to select reference points.

    *returnI* : boolean, optional
        Sets whether or not *I* is returned alongside *R* and *T*
        
        Default: False
        
    *directK* : boolean, optional
        If *directK* is true, then *K_rel* is assumed as an integer and is taken 
        as is. If *directK* is false, then *K_rel* is assumed to be a float between [0,1]
        
        Default: False
        
    *forceCalculation* : boolean, optional
        Forces calculation of reference points even in the case of full model, when 
        reference point computation is normally skipped as unnecessary
        
        Default: False
        
    *progressingI* : boolean, optional
        In the case that a lengthy reference point selection is expected, 
        *progressingI* can be used to set intermediate results and to allow 
        the computation to continue if temporary results are present should a 
        restart to continue computation be in order
    
        Default: False
        
    *saveInterval* : int, optional
        How many reference points to be computed before temporary results saved
    
        Default: 10000

    Returns
    -------

    *R* : numpy.ndarray
        Input data reference points

    *T* : numpy.ndarray
        Output data reference points
        
    If *returnI* is True: 
        
    *I* : numpy.ndarray
        Reference point indices

    """

    N = len(Y)
    
    tempFilenameI = 'RS-maximin_temp_I.npy'
    tempFilenameD = 'RS-maximin_temp_D.npy'

    if directK:
        K_abs = K_rel
    else:
        K_abs = int(np.round(N*(K_rel/100)))

    if K_abs == N and (not forceCalculation): # Full MLM
        R = X
        T = Y
        I = np.arange(N).reshape(-1,1)
        
    elif rs_method == "random":
        I = np.random.permutation(N)
        R = X[I[range(K_abs)]]
        T = Y[I[range(K_abs)]]
        I = I[range(K_abs)]
        I = I.reshape(-1,1)
        
    elif rs_method == "RS-maximin":
        I = np.zeros((K_abs,1),dtype=int)

        # First ref. point closest to data mean
        xm = np.mean(X,axis=0).reshape(1,-1)
        I[0] = np.argmin(pdc(xm,X))

        i = 1
        fill_rest = False
        D = np.full(N,float("inf"))
        if progressingI:
            #print('Check for previous files...')
            # There may be previous progress
            pI = Path(tempFilenameI)
            pD = Path(tempFilenameD)
            if pI.exists() and pD.exists():
                #print('Previous found.')
                # Savefile detected.
                I = np.load(pI)
                D = np.load(pD)

        counter = 0
        while i < K_abs:
            D2 = pdc(X,X[I[i-1]]).reshape(1,-1)
            D = np.min(np.vstack((D,D2)),axis=0)
            Id = np.where(D > 0)
            Id = Id[0]
            if len(Id) > 0:
                maxi = np.argmax(D[Id])
                I[i] = Id[maxi]
                i += 1
            else:
                fill_rest = True
                break
            
            if progressingI and counter % saveInterval == 0:
                np.save(tempFilenameI,I)
                np.save(tempFilenameD,D)
            counter+=1
            
        # Add duplicate reference points
        if fill_rest:
            I_dup = np.setdiff1d(np.arange(N).reshape(-1,1),I)
            nIdup = np.shape(I_dup)[0]
            J = I_dup[(np.random.permutation(nIdup))[:(K_abs-i)]]
            I[i:,:] = J.reshape(-1,1)

        R = X[I[:,0]]
        T = Y[I[:,0]]
        
    else:
        raise NotImplementedError('Faulty or no method selected')
        
    if progressingI:
        osremove(tempFilenameI)
        osremove(tempFilenameD)

    if returnI:
        return R, T, I[:,0]
    else:
        return R, T


# Predict with MLM
def pred(B, R, T, x, cubic=True):
    """
    Computes MLM prediction y for given input x. 
    Order of parameters is so that a partial function may be applied. 


    Parameters
    ----------

    *B* : numpy.ndarray (B_hat)
        Distance regression model coefficients.

    *R* : numpy.ndarray
        Input data reference points

    *T* : numpy.ndarray
        Output data reference points

    *x* : numpy.ndarray
        Input vector

    *cubic* : boolean, optional
        If true, cubic multilateration solver is used. If false, LLS is used instead. 
        
        Default: True

    Returns
    -------
    *y_hat* : numpy.ndarray
        Predicted output.
    """

    xshp = x.shape
    try:
        assert xshp[0] == 1
        assert xshp[1] > 1
    except AssertionError as e:
        print('xshp:',xshp)
        raise AssertionError(e)

    # Predict euclidean distances to T in output space
    delta_hat = np.matmul(pdc(x,R),B).reshape(-1,1)

    # Solve multilateration problem
    if cubic:
        y_hat = cubic_eq_solver(T, delta_hat)
    else:
        y_hat = LLS(T, delta_hat)

    return np.real(y_hat)[0,0]


# Analytical Multilateration Solver for the MLM
def cubic_eq_solver(T, delta_hat):
    """
    Solves multilateration problem as a cubic equation using np.roots
    function. See equation (11) and algorithm 2 in Mesquita et al. "Ensemble
    of Efficient Minimal Learning Machines for Classification and Regression"
    2017.

    Parameters
    ----------
    *T* : numpy.ndarray
        Output data reference points

    *delta_hat* : numpy.ndarray
        Predicted euclidean distances to reference points in output space.

    Returns
    -------
    *y_hat* : numpy.ndarray
        Root corresponding to smallest cost function value.
    """
   
    delta_hat2 = delta_hat**2

    # Third order polynomial coefficients for the deritative of the cost
    # function (11)
    A = np.size(delta_hat)
    #B = -3*np.sum(T) # original
    B = -3*np.sum(T, axis=0)
    #C = np.sum(3*(T**2)-delta_hat2) # original
    C = np.sum(3*(T**2)-delta_hat2, axis=0)
    #D = np.sum(delta_hat2*T-T**3) # original
    D = np.sum(delta_hat2*T-T**3, axis=0)

    # Remember: if T.shape == (5,3), then T[:,0].shape == (5,) !
    # One would expect the result as (5,1) bit it isn't so.
    N = B.shape[0]
    assert N == C.shape[0] and N == D.shape[0]
    #res = np.zeros(1,N)
    inp = zip(repeat(A),B,C,D,repeat(delta_hat2),repeat(T))
    res = np.array(list(map(lambda x: _cubic_helper(x), inp)))
    return res.reshape(1,-1)
    """
    # Original here in a comment.
    # Find three roots
    y = np.roots([A,B,C,D]) # Each is expected to be a single number
    I = np.array([np.isreal(y[0]), np.isreal(y[1]), np.isreal(y[2])])

    # Select a root corresponding to smallest cost function value
    Jy = np.full(3,float("inf"))
    if y[0] != y[1] and I[0] and I[1]:
        for j in range(3):
            Jy[j] = np.sum(((y[j]-T)**2-delta_hat2)**2)
        ind = np.argmin(Jy)
        y_hat = y[ind]
    else:
        y_hat = y[I]
        y_hat = y_hat[0]
    return y_hat
    """


def _cubic_helper(inp):
    """
    Multilateration solver
    """
    (A,B,C,D,dh2,T) = inp
    y = np.roots([A,B,C,D]) # Each is expected to be a single number
    I = np.array([np.isreal(y[0]), np.isreal(y[1]), np.isreal(y[2])])

    # Select a root corresponding to smallest cost function value
    Jy = np.full(3,float("inf"))
    if y[0] != y[1] and I[0] and I[1]:
        for j in range(3):
            Jy[j] = np.sum(((y[j]-T)**2-dh2)**2)
        ind = np.argmin(Jy)
        y_hat = y[ind]
    else:
        y_hat = y[I]
        y_hat = y_hat[0]
    return y_hat


def LLS(T, delta_hat):
    """
    Multilateration solver
    """
    (K,S) = T.shape
    a = np.arange(K)
    random.shuffle(a)
    ind = a[0]
    r = T[ind]
    dr = delta_hat[ind][0]
    ta = np.arange(K)
    T = T[ta!=ind]
    dha = np.arange(len(delta_hat))
    delta_hat = delta_hat[dha!=ind]
    K -=1

    A = T-r
    d_ir = pdc(T.reshape(-1,1),r.reshape(-1,1))
    b = 0.5*(np.square(dr.reshape(-1,1))+np.square(d_ir)-np.square(delta_hat.reshape(-1,1)))
    theta = np.linalg.lstsq(A.reshape(-1,1), b, rcond=None)[0]
    return (theta.T + r)[0][0]



